"""Module with group plugins to represent pseudo potential families."""
from .cutoffs import *
from .pseudo import *
from .pseudo_dojo import *
from .sssp import *

__all__ = cutoffs.__all__ + pseudo.__all__ + pseudo_dojo.__all__ + sssp.__all__



"""Custom parameter types for command line interface commands."""
from __future__ import annotations

import pathlib

import click
import requests
from aiida.cmdline.params.types import GroupParamType

from ..utils import attempt

__all__ = ('PseudoPotentialFamilyTypeParam', 'PseudoPotentialFamilyParam', 'PseudoPotentialTypeParam')


class PseudoPotentialTypeParam(click.ParamType):
    """Parameter type for ``click`` commands to define a subclass of ``PseudoPotentialData``."""

    name = 'pseudo_type'

    def convert(self, value, _, __):
        """Convert the entry point name to the corresponding class.

        :param value: entry point name that should correspond to subclass of ``PseudoPotentialData`` data plugin
        :return: the ``PseudoPotentialData`` subclass
        :raises: ``click.BadParameter`` if the entry point cannot be loaded or is not subclass of
            ``PseudoPotentialData``
        """
        from aiida.common import exceptions
        from aiida.plugins import DataFactory

        from aiida_pseudo.data.pseudo import PseudoPotentialData

        try:
            pseudo_type = DataFactory(value)
        except exceptions.EntryPointError as exception:
            raise click.BadParameter(f'`{value}` is not an existing data plugin.') from exception

        if not issubclass(pseudo_type, PseudoPotentialData):
            raise click.BadParameter(f'`{value}` entry point is not a subclass of `PseudoPotentialData`.')

        PseudoPotentialData.entry_point = value

        return pseudo_type

    def complete(self, _, incomplete):
        """Return possible completions based on an incomplete value.

        :returns: list of tuples of valid entry points (matching incomplete) and a description
        """
        from aiida.plugins.entry_point import get_entry_point_names

        entry_points = get_entry_point_names('aiida.data')
        return [(ep, '') for ep in entry_points if (ep.startswith('pseudo') and ep.startswith(incomplete))]


class PseudoPotentialFamilyParam(GroupParamType):
    """Parameter type for ``click`` commands to define an instance of a ``PseudoPotentialFamily``."""

    name = 'pseudo_family'

    def __init__(self, blacklist: list[str] | None = None, **kwargs):
        """Construct the parameter.

        :param blacklist: an optional list of values that should be considered invalid and will raise ``BadParameter``.
        """
        super().__init__(**kwargs)
        self.blacklist = blacklist

    def convert(self, value, param, ctx):
        """Convert the entry point name to the corresponding class.

        :param value: entry point name that should correspond to subclass of ``PseudoPotentialFamily`` group plugin
        :return: the ``PseudoPotentialFamily`` subclass
        :raises: `click.BadParameter` if the entry point cannot be loaded or is not subclass of `PseudoPotentialFamily`
        """
        family = super().convert(value, param, ctx)

        if self.blacklist and family.type_string in self.blacklist:
            self.fail(f'The value `{family}` is not allowed for this parameter.', param, ctx)
        return family


class PseudoPotentialFamilyTypeParam(click.ParamType):
    """Parameter type for ``click`` commands to define a subclass of ``PseudoPotentialFamily``."""

    name = 'pseudo_family_type'

    def __init__(self, blacklist: list[str] | None = None, **kwargs):
        """Construct the parameter.

        :param blacklist: an optional list of values that should be considered invalid and will raise ``BadParameter``.
        """
        super().__init__(**kwargs)
        self.blacklist = blacklist

    def convert(self, value, _, __):
        """Convert the entry point name to the corresponding class.

        :param value: entry point name that should correspond to subclass of ``PseudoPotentialFamily`` group plugin
        :return: the ``PseudoPotentialFamily`` subclass
        :raises: `click.BadParameter` if the entry point cannot be loaded or is not subclass of `
            `PseudoPotentialFamily``
        """
        from aiida.common import exceptions
        from aiida.plugins import GroupFactory

        from aiida_pseudo.groups.family import PseudoPotentialFamily

        try:
            family_type = GroupFactory(value)
        except exceptions.EntryPointError as exception:
            raise click.BadParameter(f'`{value}` is not an existing group plugin.') from exception

        if self.blacklist and value in self.blacklist:
            raise click.BadParameter(f'`{value}` is not an accepted value for this option.')

        if not issubclass(family_type, PseudoPotentialFamily):
            raise click.BadParameter(f'`{value}` entry point is not a subclass of `PseudoPotentialFamily`.')

        PseudoPotentialFamily.entry_point = value

        return family_type

    def complete(self, _, incomplete):
        """Return possible completions based on an incomplete value.

        :returns: list of tuples of valid entry points (matching incomplete) and a description
        """
        from aiida.plugins.entry_point import get_entry_point_names

        entry_points = get_entry_point_names('aiida.groups')
        return [(ep, '') for ep in entry_points if (ep.startswith('pseudo.family') and ep.startswith(incomplete))]


class PathOrUrl(click.Path):
    """Extension of ``click``'s ``Path``-type that also supports URLs."""

    name = 'PathOrUrl'

    def convert(self, value, param, ctx) -> pathlib.Path | bytes:
        """Convert the string value to the desired value.

        If the ``value`` corresponds to a valid path on the local filesystem, return it as a ``pathlib.Path`` instance.
        Otherwise, treat it as a URL and try to fetch the content. If successful, the raw retrieved bytes will be
        returned.

        :param value: the filepath on the local filesystem or a URL.
        """
        try:
            # Call the method of the super class, which will raise if it ``value`` is not a valid path.
            return pathlib.Path(super().convert(value, param, ctx))
        except click.exceptions.BadParameter:
            with attempt(f'attempting to download data from `{value}`...'):
                response = requests.get(value, timeout=30)
                response.raise_for_status()
                return response


class UnitParamType(click.ParamType):
    """Parameter type to specify units from the ``pint`` library."""

    name = 'unit'

    def __init__(self, quantity: list[str] | None = None, **kwargs):
        """Construct the parameter.

        :param quantity: The corresponding quantity of the unit.
        """
        super().__init__(**kwargs)
        self.quantity = quantity

    def convert(self, value, _, __):
        """Check if the provided unit is a valid unit for the defined quantity.

        :raises: ``click.BadParameter`` if the provided unit is not valid for the quantity defined for this instance.
        """
        from ...common.units import U

        if value not in U:
            raise click.BadParameter(f'`{value}` is not a valid unit.')

        if not U.Quantity(1, value).check(f'[{self.quantity}]'):
            raise click.BadParameter(f'`{value}` is not a valid `{self.quantity}` unit.')

        return value



"""Module with constants for unit conversions."""
from pint import UnitRegistry

# This unit registry singleton should be used to construct new quantities with a unit and to convert them to other units
U = UnitRegistry()



"""Command line interface utilities."""
from contextlib import contextmanager
from pathlib import Path

from aiida.cmdline.utils import echo

__all__ = ('attempt', 'create_family_from_archive')


@contextmanager
def attempt(message, exception_types=Exception, include_traceback=False):
    """Context manager to be used to wrap statements in CLI that can throw exceptions.

    :param message: the message to print before yielding
    :param include_traceback: boolean, if True, will also print traceback if an exception is caught
    """
    import sys
    import traceback

    echo.echo_report(message, nl=False)

    try:
        yield
    except exception_types as exception:
        echo.echo(' [FAILED]', fg='red', bold=True)
        message = str(exception)
        if include_traceback:
            message += f"\n{''.join(traceback.format_exception(*sys.exc_info()))}"
        echo.echo_critical(message)
    else:
        echo.echo(' [OK]', fg='green', bold=True)


def create_family_from_archive(cls, label, filepath_archive: Path, fmt=None, pseudo_type=None):
    """Construct a new pseudo family instance from a tar.gz archive.

    .. warning:: the archive should not contain any subdirectories, but just the pseudo potential files.

    :param cls: the pseudopotential family class to use, e.g. ``SsspFamily``
    :param label: the label for the new family
    :param filepath_archive: absolute filepath to the .tar.gz archive containing the pseudo potentials
    :param fmt: the format of the archive, if not specified will attempt to guess based on extension of ``filepath``
    :param pseudo_type: subclass of ``PseudoPotentialData`` to be used for the parsed pseudos. If not specified and
        the family only defines a single supported pseudo type in ``_pseudo_types`` then that will be used otherwise
        a ``ValueError`` is raised.
    :return: newly created family
    :raises OSError: if the archive could not be unpacked or pseudos in it could not be parsed into a family
    """
    import shutil
    import tempfile

    with tempfile.TemporaryDirectory() as dirpath:
        try:
            shutil.unpack_archive(filepath_archive, dirpath, format=fmt)
        except shutil.ReadError as exception:
            raise OSError(f'failed to unpack the archive `{filepath_archive}`: {exception}') from exception

        try:
            family = cls.create_from_folder(Path(dirpath), label, pseudo_type=pseudo_type)
        except ValueError as exception:
            raise OSError(f'failed to parse pseudos from `{dirpath}`: {exception}') from exception

    return family



"""Module for data plugin to represent a pseudo potential in UPF format."""
import pathlib
import re
import typing

from .pseudo import PseudoPotentialData

__all__ = ('UpfData',)

REGEX_ELEMENT_V1 = re.compile(r"""(?P<element>[a-zA-Z]{1,2})\s+Element""")
REGEX_ELEMENT_V2 = re.compile(r"""\s*element\s*=\s*['"]\s*(?P<element>[a-zA-Z]{1,2})\s*['"].*""")

PATTERN_FLOAT = r'[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?'
REGEX_Z_VALENCE_V1 = re.compile(r"""(?P<z_valence>""" + PATTERN_FLOAT + r""")\s+Z valence""")
REGEX_Z_VALENCE_V2 = re.compile(r"""\s*z_valence\s*=\s*['"]\s*(?P<z_valence>""" + PATTERN_FLOAT + r""")\s*['"].*""")


def parse_element(content: str):
    """Parse the content of the UPF file to determine the element.

    :param stream: a filelike object with the binary content of the file.
    :return: the symbol of the element following the IUPAC naming standard.
    """
    for regex in [REGEX_ELEMENT_V2, REGEX_ELEMENT_V1]:
        match = regex.search(content)

        if match:
            return match.group('element')

    raise ValueError(f'could not parse the element from the UPF content: {content}')


def parse_z_valence(content: str) -> int:
    """Parse the content of the UPF file to determine the Z valence.

    :param stream: a filelike object with the binary content of the file.
    :return: the Z valence.
    """
    for regex in [REGEX_Z_VALENCE_V2, REGEX_Z_VALENCE_V1]:
        match = regex.search(content)

        if match:
            z_valence = match.group('z_valence')

            try:
                z_valence = float(z_valence)
            except ValueError as exception:
                raise ValueError(f'parsed value for the Z valence `{z_valence}` is not a valid number.') from exception

            if int(z_valence) != z_valence:
                raise ValueError(f'parsed value for the Z valence `{z_valence}` is not an integer.')

            return int(z_valence)

    raise ValueError(f'could not parse the Z valence from the UPF content: {content}')


class UpfData(PseudoPotentialData):
    """Data plugin to represent a pseudo potential in UPF format."""

    _key_z_valence = 'z_valence'

    def set_file(
        self, source: typing.Union[str, pathlib.Path, typing.BinaryIO], filename: typing.Optional[str] = None, **kwargs
    ):
        """Set the file content and parse other optional attributes from the content.

        .. note:: this method will first analyse the type of the ``source`` and if it is a filepath will convert it
            to a binary stream of the content located at that filepath, which is then passed on to the superclass. This
            needs to be done first, because it will properly set the file and filename attributes that are expected by
            other methods. Straight after the superclass call, the source seeker needs to be reset to zero if it needs
            to be read again, because the superclass most likely will have read the stream to the end. Finally it is
            important that the ``prepare_source`` is called here before the superclass invocation, because this way the
            conversion from filepath to byte stream will be performed only once. Otherwise, each subclass would perform
            the conversion over and over again.

        :param source: the source pseudopotential content, either a binary stream, or a ``str`` or ``Path`` to the path
            of the file on disk, which can be relative or absolute.
        :param filename: optional explicit filename to give to the file stored in the repository.
        :raises TypeError: if the source is not a ``str``, ``pathlib.Path`` instance or binary stream.
        :raises FileNotFoundError: if the source is a filepath but does not exist.
        :raises ValueError: if the element symbol is invalid.
        """
        source = self.prepare_source(source)
        super().set_file(source, filename, **kwargs)
        source.seek(0)
        content = source.read().decode('utf-8')
        self.element = parse_element(content)
        self.z_valence = parse_z_valence(content)

    @property
    def z_valence(self) -> typing.Optional[int]:
        """Return the Z valence.

        :return: the Z valence.
        """
        return self.base.attributes.get(self._key_z_valence, None)

    @z_valence.setter
    def z_valence(self, value: int):
        """Set the Z valence.

        :param value: the Z valence.
        :raises ValueError: if the value is not a positive integer.
        """
        if not isinstance(value, int) or value < 0:
            raise ValueError(f'`{value}` is not a positive integer')

        self.base.attributes.set(self._key_z_valence, value)



"""Subclass of `PseudoPotentialFamily` designed to represent a PseudoDojo configuration."""
from __future__ import annotations

import json
import pathlib
import re
import warnings
from typing import ClassVar, NamedTuple, Sequence

from aiida.common.exceptions import ParsingError

from aiida_pseudo.data.pseudo import JthXmlData, PsmlData, Psp8Data, UpfData

from ..mixins import RecommendedCutoffMixin
from .pseudo import PseudoPotentialFamily

__all__ = ('PseudoDojoConfiguration', 'PseudoDojoFamily')


class PseudoDojoConfiguration(NamedTuple):
    """Named tuple that represents a PseudoDojo configuration."""

    version: str
    functional: str
    relativistic: str
    protocol: str
    pseudo_format: str

    def __str__(self):
        """Represent the configuration as a string."""
        return f'PseudoDojo v{self.version} {self.functional} {self.relativistic} {self.protocol} {self.pseudo_format}'


class PseudoDojoFamily(RecommendedCutoffMixin, PseudoPotentialFamily):
    """Subclass of ``PseudoPotentialFamily`` designed to represent a PseudoDojo configuration.

    The ``PseudoDojoFamily`` is essentially a ``PseudoPotentialFamily`` with some additional constraints. It can only
    be used to contain the pseudo potentials and corresponding metadata of an official PseudoDojo configuration.
    """

    _pseudo_types = (UpfData, PsmlData, Psp8Data, JthXmlData)
    _pseudo_repo_file_extensions = ('djrepo',)

    label_template = 'PseudoDojo/{version}/{functional}/{relativistic}/{protocol}/{pseudo_format}'
    default_configuration = PseudoDojoConfiguration('0.4', 'PBE', 'SR', 'standard', 'psp8')
    valid_configurations = (
        PseudoDojoConfiguration('0.4', 'PBE', 'SR', 'standard', 'psp8'),
        PseudoDojoConfiguration('0.4', 'PBE', 'SR', 'stringent', 'psp8'),
        PseudoDojoConfiguration('0.4', 'PBEsol', 'SR', 'standard', 'psp8'),
        PseudoDojoConfiguration('0.4', 'PBEsol', 'SR', 'stringent', 'psp8'),
        PseudoDojoConfiguration('0.4', 'LDA', 'SR', 'standard', 'psp8'),
        PseudoDojoConfiguration('0.4', 'LDA', 'SR', 'stringent', 'psp8'),
        PseudoDojoConfiguration('0.4', 'PBE', 'SR3plus', 'standard', 'psp8'),
        PseudoDojoConfiguration('0.4', 'PBE', 'FR', 'standard', 'psp8'),
        PseudoDojoConfiguration('0.4', 'PBE', 'FR', 'stringent', 'psp8'),
        PseudoDojoConfiguration('0.4', 'PBEsol', 'FR', 'standard', 'psp8'),
        PseudoDojoConfiguration('0.4', 'PBEsol', 'FR', 'stringent', 'psp8'),
        PseudoDojoConfiguration('0.4', 'PBE', 'SR', 'standard', 'upf'),
        PseudoDojoConfiguration('0.4', 'PBE', 'SR', 'stringent', 'upf'),
        PseudoDojoConfiguration('0.4', 'PBEsol', 'SR', 'standard', 'upf'),
        PseudoDojoConfiguration('0.4', 'PBEsol', 'SR', 'stringent', 'upf'),
        PseudoDojoConfiguration('0.4', 'LDA', 'SR', 'standard', 'upf'),
        PseudoDojoConfiguration('0.4', 'LDA', 'SR', 'stringent', 'upf'),
        PseudoDojoConfiguration('0.4', 'PBE', 'SR3plus', 'standard', 'upf'),
        PseudoDojoConfiguration('0.4', 'PBE', 'FR', 'standard', 'upf'),
        PseudoDojoConfiguration('0.4', 'PBE', 'FR', 'stringent', 'upf'),
        PseudoDojoConfiguration('0.4', 'PBEsol', 'FR', 'standard', 'upf'),
        PseudoDojoConfiguration('0.4', 'PBEsol', 'FR', 'stringent', 'upf'),
        PseudoDojoConfiguration('0.4', 'PBE', 'SR', 'standard', 'psml'),
        PseudoDojoConfiguration('0.4', 'PBE', 'SR', 'stringent', 'psml'),
        PseudoDojoConfiguration('0.4', 'PBEsol', 'SR', 'standard', 'psml'),
        PseudoDojoConfiguration('0.4', 'PBEsol', 'SR', 'stringent', 'psml'),
        PseudoDojoConfiguration('0.4', 'LDA', 'SR', 'standard', 'psml'),
        PseudoDojoConfiguration('0.4', 'LDA', 'SR', 'stringent', 'psml'),
        PseudoDojoConfiguration('0.4', 'PBE', 'SR3plus', 'standard', 'psml'),
        PseudoDojoConfiguration('0.4', 'PBE', 'FR', 'standard', 'psml'),
        PseudoDojoConfiguration('0.4', 'PBE', 'FR', 'stringent', 'psml'),
        PseudoDojoConfiguration('0.4', 'PBEsol', 'FR', 'standard', 'psml'),
        PseudoDojoConfiguration('0.4', 'PBEsol', 'FR', 'stringent', 'psml'),
        PseudoDojoConfiguration('0.5', 'PBE', 'SR', 'standard', 'psp8'),
        PseudoDojoConfiguration('0.5', 'PBE', 'SR', 'stringent', 'psp8'),
        PseudoDojoConfiguration('0.5', 'PBE', 'SR', 'standard', 'upf'),
        PseudoDojoConfiguration('0.5', 'PBE', 'SR', 'stringent', 'upf'),
        PseudoDojoConfiguration('0.5', 'PBE', 'SR', 'standard', 'psml'),
        PseudoDojoConfiguration('0.5', 'PBE', 'SR', 'stringent', 'psml'),
        PseudoDojoConfiguration('1.0', 'PBE', 'SR', 'standard', 'jthxml'),
        PseudoDojoConfiguration('1.0', 'PBE', 'SR', 'stringent', 'jthxml'),
        PseudoDojoConfiguration('1.0', 'LDA', 'SR', 'standard', 'jthxml'),
        PseudoDojoConfiguration('1.0', 'LDA', 'SR', 'stringent', 'jthxml'),
    )

    url_base = 'http://www.pseudo-dojo.org/pseudos/'
    urls: ClassVar[dict[str, str]] = {
        'PseudoDojo/0.4/PBE/SR/standard/psp8': 'nc-sr-04_pbe_standard_psp8',
        'PseudoDojo/0.4/PBE/SR/stringent/psp8': 'nc-sr-04_pbe_stringent_psp8',
        'PseudoDojo/0.4/PBEsol/SR/standard/psp8': 'nc-sr-04_pbesol_standard_psp8',
        'PseudoDojo/0.4/PBEsol/SR/stringent/psp8': 'nc-sr-04_pbesol_stringent_psp8',
        'PseudoDojo/0.4/LDA/SR/standard/psp8': 'nc-sr-04_pw_standard_psp8',
        'PseudoDojo/0.4/LDA/SR/stringent/psp8': 'nc-sr-04_pw_stringent_psp8',
        'PseudoDojo/0.4/PBE/SR3plus/standard/psp8': 'nc-sr-04-3plus_pbe_standard_psp8',
        'PseudoDojo/0.4/PBE/FR/standard/psp8': 'nc-fr-04_pbe_standard_psp8',
        'PseudoDojo/0.4/PBE/FR/stringent/psp8': 'nc-fr-04_pbe_stringent_psp8',
        'PseudoDojo/0.4/PBEsol/FR/standard/psp8': 'nc-fr-04_pbesol_standard_psp8',
        'PseudoDojo/0.4/PBEsol/FR/stringent/psp8': 'nc-fr-04_pbesol_stringent_psp8',
        'PseudoDojo/0.4/PBE/SR/standard/upf': 'nc-sr-04_pbe_standard_upf',
        'PseudoDojo/0.4/PBE/SR/stringent/upf': 'nc-sr-04_pbe_stringent_upf',
        'PseudoDojo/0.4/PBEsol/SR/standard/upf': 'nc-sr-04_pbesol_standard_upf',
        'PseudoDojo/0.4/PBEsol/SR/stringent/upf': 'nc-sr-04_pbesol_stringent_upf',
        'PseudoDojo/0.4/LDA/SR/standard/upf': 'nc-sr-04_pw_standard_upf',
        'PseudoDojo/0.4/LDA/SR/stringent/upf': 'nc-sr-04_pw_stringent_upf',
        'PseudoDojo/0.4/PBE/SR3plus/standard/upf': 'nc-sr-04-3plus_pbe_standard_upf',
        'PseudoDojo/0.4/PBE/FR/standard/upf': 'nc-fr-04_pbe_standard_upf',
        'PseudoDojo/0.4/PBE/FR/stringent/upf': 'nc-fr-04_pbe_stringent_upf',
        'PseudoDojo/0.4/PBEsol/FR/standard/upf': 'nc-fr-04_pbesol_standard_upf',
        'PseudoDojo/0.4/PBEsol/FR/stringent/upf': 'nc-fr-04_pbesol_stringent_upf',
        'PseudoDojo/0.4/PBE/SR/standard/psml': 'nc-sr-04_pbe_standard_psml',
        'PseudoDojo/0.4/PBE/SR/stringent/psml': 'nc-sr-04_pbe_stringent_psml',
        'PseudoDojo/0.4/PBEsol/SR/standard/psml': 'nc-sr-04_pbesol_standard_psml',
        'PseudoDojo/0.4/PBEsol/SR/stringent/psml': 'nc-sr-04_pbesol_stringent_psml',
        'PseudoDojo/0.4/LDA/SR/standard/psml': 'nc-sr-04_pw_standard_psml',
        'PseudoDojo/0.4/LDA/SR/stringent/psml': 'nc-sr-04_pw_stringent_psml',
        'PseudoDojo/0.4/PBE/SR3plus/standard/psml': 'nc-sr-04-3plus_pbe_standard_psml',
        'PseudoDojo/0.4/PBE/FR/standard/psml': 'nc-fr-04_pbe_standard_psml',
        'PseudoDojo/0.4/PBE/FR/stringent/psml': 'nc-fr-04_pbe_stringent_psml',
        'PseudoDojo/0.4/PBEsol/FR/standard/psml': 'nc-fr-04_pbesol_standard_psml',
        'PseudoDojo/0.4/PBEsol/FR/stringent/psml': 'nc-fr-04_pbesol_stringent_psml',
        'PseudoDojo/0.5/PBE/SR/standard/psp8': 'nc-sr-05_pbe_standard_psp8',
        'PseudoDojo/0.5/PBE/SR/stringent/psp8': 'nc-sr-05_pbe_stringent_psp8',
        'PseudoDojo/0.5/PBE/SR/standard/upf': 'nc-sr-05_pbe_standard_upf',
        'PseudoDojo/0.5/PBE/SR/stringent/upf': 'nc-sr-05_pbe_stringent_upf',
        'PseudoDojo/0.5/PBE/SR/standard/psml': 'nc-sr-05_pbe_standard_psml',
        'PseudoDojo/0.5/PBE/SR/stringent/psml': 'nc-sr-05_pbe_stringent_psml',
        'PseudoDojo/1.0/PBE/SR/standard/jthxml': 'paw_pbe_standard_xml',
        'PseudoDojo/1.0/PBE/SR/stringent/jthxml': 'paw_pbe_stringent_xml',
        'PseudoDojo/1.0/LDA/SR/standard/jthxml': 'paw_pw_standard_xml',
        'PseudoDojo/1.0/LDA/SR/stringent/jthxml': 'paw_pw_stringent_xml',
    }

    @classmethod
    def get_valid_labels(cls) -> Sequence[str]:
        """Return the tuple of labels of all valid PseudoDojo configurations."""
        configurations = set(cls.valid_configurations)
        return tuple(cls.format_configuration_label(configuration) for configuration in configurations)

    @classmethod
    def format_configuration_label(cls, configuration: PseudoDojoConfiguration) -> str:
        """Format a label for an `PseudoDojoFamily` with the required syntax.

        :param configuration: the PseudoDojo configuration.
        :return: label.
        """
        return cls.label_template.format(
            version=configuration.version,
            functional=configuration.functional,
            relativistic=configuration.relativistic,
            protocol=configuration.protocol,
            pseudo_format=configuration.pseudo_format,
        )

    @classmethod
    def get_url_archive(cls, label):
        """Return the URL for the pseudopotential archive file for a given family label.

        :param label: the family label as formatted by ``format_configuration_label``.
        :return: the URL from which the pseudopotential archive can be downloaded.
        :raises: `ValueError` if the label is not a valid configuration label.
        """
        try:
            url_prefix = cls.urls[label]
        except KeyError as exception:
            raise ValueError(f'label {label} is not a valid PseudoDojo configuration label.') from exception

        return cls.url_base + url_prefix + '.tgz'

    @classmethod
    def get_url_metadata(cls, label):
        """Return the URL for the pseudopotential metadata file for a given family label.

        :param label: the family label as formatted by ``format_configuration_label``.
        :return: the URL from which the pseudopotential metadata can be downloaded.
        :raises: `ValueError` if the label is not a valid configuration label.
        """
        try:
            url_prefix = cls.urls[label]
        except KeyError as exception:
            raise ValueError(f'label {label} is not a valid PseudoDojo configuration label.') from exception

        # For the metadata URL, the final underscored part of the prefix (the pseudo format) is replaced with `_djrepo`.
        url_prefix = '_'.join(url_prefix.split('_')[:-1] + ['djrepo'])

        return cls.url_base + url_prefix + '.tgz'

    @classmethod
    def get_md5_from_djrepo(cls, djrepo, pseudo_type):
        """Get the appropriate md5 hash from a DJREPO file.

        :param djrepo: dictionary loaded from DJREPO JSON file.
        :reutnrs: md5 string.
        """
        md5_key_mapping = {UpfData: 'md5_upf', Psp8Data: 'md5', PsmlData: 'md5_psml', JthXmlData: 'md5'}

        try:
            md5_key = md5_key_mapping[pseudo_type]
        except KeyError as exception:
            raise ValueError(
                f'pseudo type `{pseudo_type}` is unsupported by PseudoDojo djrepos: {exception}'
            ) from exception

        try:
            md5 = djrepo[md5_key]
        except KeyError as exception:
            raise ParsingError(f'key `{md5_key}` is not defined in the djrepo: {exception}') from exception

        return md5

    @classmethod
    def get_cutoffs_from_djrepo(cls, djrepo, pseudo_type):
        """Collect and organize the suggested cutoffs (hints) from a DJREPO file.

        DJREPO files only provide a kinetic energy cutoff, so for pseudo types which contain norm-conserving pseudos
        from PseudoDojo, we use a dual of 8.0 to generate the charge density (rho) cutoff.

        For PAW potentials from PseudoDojo (which are assumed to be JthXmlData), a dual of 2.0 is used.

        The cutoffs in DJREPO files are given in Hartree, which is converted to eV.

        :param djrepo: dictionary loaded from DJREPO JSON file
        :returns: cutoffs dictionary (in eV) where keys are stringency levels and values are
            {'cutoff_wfc': ..., 'cutoff_rho': ...}
        """
        dual_mapping = {UpfData: 4.0, Psp8Data: 4.0, PsmlData: 4.0, JthXmlData: 2.0}

        try:
            dual = dual_mapping[pseudo_type]
        except KeyError as exception:
            raise ValueError(
                f'cannot get cutoffs for pseudo type `{pseudo_type}` because the appropriate dual '
                'for generating density cutoffs is unknown'
            ) from exception

        cutoffs = {}

        try:
            hints = djrepo.get('hints', None) or djrepo['ppgen_hints']
        except KeyError as exception:
            raise ParsingError('neither `hints` or `ppgen_hints` are defined in the djrepo.') from exception

        for stringency in ['low', 'normal', 'high']:
            try:
                ecutwfc = hints.get(stringency, {})['ecut']
            except KeyError as exception:
                raise ParsingError(f'stringency `{stringency}` is not defined in the djrepo `hints`') from exception

            cutoffs[stringency] = {'cutoff_wfc': ecutwfc, 'cutoff_rho': ecutwfc * dual}

        return cutoffs

    @classmethod
    def parse_djrepos_from_folder(cls, dirpath: pathlib.Path, pseudo_type):
        """Parse the djrepo files in the given directory into a list of data nodes.

        .. note:: The directory pointed to by `dirpath` should only contain djrepo files. Optionally, it can contain
            just a single directory, that contains all the djrepo files. If any other files are stored in the basepath
            or the subdirectory that cannot be successfully parsed as djrepo files the method will raise a `ValueError`.

        :param dirpath: absolute path to a directory containing djrepos.
        :return: list of data nodes.
        :raises ValueError: if `dirpath` is not a directory or contains anything other than files.
        :raises ValueError: if `dirpath` contains multiple djrepos for the same element.
        :raises ParsingError: if the constructor of the pseudo type fails for one of the files in the `dirpath`.
        """
        md5s = {}
        cutoffs = {'low': {}, 'normal': {}, 'high': {}}
        elements = []

        dirpath = cls._validate_dirpath(dirpath)

        for filepath in dirpath.iterdir():
            filename = filepath.name

            if not filepath.is_file():
                raise ValueError(f'dirpath `{dirpath}` contains at least one entry that is not a file: {filepath}')

            # Some of the djrepo archives contain extraneous files. Here we skip files with unsupported extensions.
            if filename.split('.')[-1] not in cls._pseudo_repo_file_extensions:
                warnings.warn(f'filename {filename} does not have a supported extension. Skipping...')
                continue

            try:
                with open(filepath, 'r', encoding='utf-8') as handle:
                    djrepo = json.load(handle)
            except ParsingError as exception:
                raise ParsingError(f'failed to parse `{filepath}`: {exception}') from exception
            else:
                match = re.search(r'^([A-Za-z]{1,2})\.\w+', filename)
                if match is None:
                    raise ParsingError(
                        f'could not parse a valid element symbol from the filename `{filename}`. '
                        'It should have the format `ELEMENT.EXTENSION`'
                    )
                element = match.group(1)
                if element in elements:
                    raise ValueError(f'directory `{dirpath}` contains djrepos with duplicate elements`')

                try:
                    md5 = cls.get_md5_from_djrepo(djrepo, pseudo_type=pseudo_type)
                except (ParsingError, ValueError) as exception:
                    raise ParsingError(f'failed to parse md5 from djrepo file `{filename}`: {exception}') from exception
                else:
                    md5s[element] = md5

                try:
                    djrepo_cutoffs = cls.get_cutoffs_from_djrepo(djrepo, pseudo_type=pseudo_type)
                except ParsingError as exception:
                    raise ParsingError(
                        f'failed to parse cutoffs from djrepo file `{filename}`: {exception}'
                    ) from exception
                else:
                    for stringency in ['low', 'normal', 'high']:
                        cutoffs[stringency][element] = djrepo_cutoffs[stringency]

                elements.append(element)

        if (not cutoffs['low']) and (not cutoffs['normal']) and (not cutoffs['high']):
            raise ValueError(f'no djrepos were parsed from `{dirpath}`')

        return md5s, cutoffs

    @classmethod
    def parse_djrepos_from_archive(cls, filepath_metadata: pathlib.Path, fmt=None, pseudo_type=None):
        """Parse metadata from a djrepo .tgz archive.

        .. warning:: the archive should not contain any subdirectories, but just the djrepo files.

        :param filepath_metadata: the path to the metadata .tgz archive.
        :param fmt: the format of the archive, if not specified will attempt to guess based on extension of `filepath`.
        :param pseudo_type: subclass of ``PseudoPotentialData`` to be used for the parsed pseudos. If not specified and
            the family only defines a single supported pseudo type in ``_pseudo_types`` then that will be used otherwise
            a ``ValueError`` is raised.
        :return: element: value dictionaries containing md5s and cutoffs.
        :raises OSError: if the archive could not be unpacked or the djrepos in it could not be parsed into md5s and
            cutoffs.
        """
        import shutil
        import tempfile

        with tempfile.TemporaryDirectory() as dirpath:
            try:
                shutil.unpack_archive(filepath_metadata, dirpath, format=fmt)
            except shutil.ReadError as exception:
                raise OSError(
                    f'failed to unpack the metadata archive `{filepath_metadata}`: {exception}'
                ) from exception

            try:
                md5s, cutoffs = cls.parse_djrepos_from_folder(pathlib.Path(dirpath), pseudo_type=pseudo_type)
            except ValueError as exception:
                raise OSError(f'failed to parse djrepos from `{dirpath}`: {exception}') from exception

        return md5s, cutoffs

    def __init__(self, label=None, **kwargs):
        """Construct a new instance, validating that the label matches the required format."""
        if label not in self.get_valid_labels():
            raise ValueError(f'the label `{label}` is not a valid PseudoDojo configuration label.')

        super().__init__(label=label, **kwargs)



"""Subclass of ``Group`` that serves as a base class for representing pseudo potential families."""
import re
from typing import List, Mapping, Optional, Tuple, Union

from aiida.common import exceptions
from aiida.common.lang import classproperty, type_check
from aiida.orm import Group, QueryBuilder
from aiida.plugins import DataFactory

from aiida_pseudo.data.pseudo import PseudoPotentialData

__all__ = ('PseudoPotentialFamily',)

LegacyStructureData = DataFactory('core.structure')  # pylint: disable=invalid-name

try:
    StructureData = DataFactory('atomistic.structure')
except exceptions.MissingEntryPointError:
    structures_classes = (LegacyStructureData,)
else:
    structures_classes = (LegacyStructureData, StructureData)


class PseudoPotentialFamily(Group):
    """Group to represent a pseudo potential family.

    This is a base class that provides most of the functionality but does not actually define what type of pseudo
    potentials can be contained. If ``_pseudo_types`` is not defined, any pseudo potential type is accepted in this
    family, as long as it is a subclass of ``PseudoPotentialData``. Subclasses can limit which pseudo types can be
    hosted by setting ``_pseudo_types`` to a tuple of ``PseudoPotentialData`` subclasses.
    """

    _key_pseudo_type = '_pseudo_type'
    _pseudo_types = (PseudoPotentialData,)
    _pseudos = None

    def __repr__(self):
        """Represent the instance for debugging purposes."""
        return f'{self.__class__.__name__}<{self.pk or self.uuid}>'

    def __str__(self):
        """Represent the instance for human-readable purposes."""
        return f'{self.__class__.__name__}<{self.label}>'

    def __init__(self, *args, **kwargs):
        """Validate that the ``_pseudo_types`` class attribute is a tuple of ``PseudoPotentialData`` subclasses."""
        if (
            not self._pseudo_types
            or not isinstance(self._pseudo_types, tuple)
            or any(not issubclass(pseudo_type, PseudoPotentialData) for pseudo_type in self._pseudo_types)
        ):
            raise RuntimeError('`_pseudo_types` should be a tuple of `PseudoPotentialData` subclasses.')

        super().__init__(*args, **kwargs)

    @classproperty
    def pseudo_types(cls):  # noqa: N805
        """Return the pseudo potential types that this family accepts.

        :return: the tuple of subclasses of ``PseudoPotentialData`` that this family can host nodes of. If it returns
            ``None``, that means all subclasses are supported.
        """
        return cls._pseudo_types

    @classmethod
    def _validate_pseudo_type(cls, pseudo_type):
        """Validate the ``pseudo_type`` passed to ``parse_pseudos_from_directory``.

        :return: the pseudo type to be used.
        """
        if pseudo_type is None and len(cls._pseudo_types) > 1:
            raise ValueError(f'`{cls}` supports more than one type, so `pseudo_type` needs to be explicitly passed.')

        pseudo_type = pseudo_type or cls._pseudo_types[0]

        if all(not issubclass(pseudo_type, supported_type) for supported_type in cls._pseudo_types):
            raise ValueError(f'`{pseudo_type}` is not supported by `{cls}`.')

        return pseudo_type

    @classmethod
    def _validate_dirpath(cls, dirpath):
        """Validate the ``dirpath`` passed to ``parse_pseudos_from_directory``.

        :return: the directory path to be used.
        """
        if not dirpath.is_dir():
            raise ValueError(f'`{dirpath}` is not a directory')

        dirpath_contents = list(dirpath.iterdir())

        if len(dirpath_contents) == 1 and (dirpath / dirpath_contents[0]).is_dir():
            dirpath = dirpath_contents[0]

        return dirpath

    @classmethod
    def parse_pseudos_from_directory(cls, dirpath, pseudo_type=None, deduplicate=True):
        """Parse the pseudo potential files in the given directory into a list of data nodes.

        .. note:: The directory pointed to by `dirpath` should only contain pseudo potential files. Optionally, it can
            contain just a single directory, that contains all the pseudo potential files. If any other files are stored
            in the basepath or the subdirectory, that cannot be successfully parsed as pseudo potential files the method
            will raise a ``ValueError``.

        :param dirpath: absolute path to a directory containing pseudo potentials.
        :param pseudo_type: subclass of ``PseudoPotentialData`` to be used for the parsed pseudos. If not specified and
            the family only defines a single supported pseudo type in ``_pseudo_types`` then that will be used otherwise
            a ``ValueError`` is raised.
        :param deduplicate: if True, will scan database for existing pseudo potentials of same type and with the same
            md5 checksum, and use that instead of the parsed one.
        :return: list of data nodes
        :raises ValueError: if ``dirpath`` is not a directory or contains anything other than files.
        :raises ValueError: if ``dirpath`` contains multiple pseudo potentials for the same element.
        :raises ValueError: if ``pseudo_type`` is explicitly specified and is not supported by this family class.
        :raises ValueError: if ``pseudo_type`` is not specified and the class supports more than one pseudo type.
        :raises ParsingError: if the constructor of the pseudo type fails for one of the files in the ``dirpath``.
        """
        from aiida.common.exceptions import ParsingError

        pseudos = []
        dirpath = cls._validate_dirpath(dirpath)
        pseudo_type = cls._validate_pseudo_type(pseudo_type)

        for filepath in dirpath.iterdir():
            filename = filepath.name

            if not filepath.is_file():
                raise ValueError(f'dirpath `{dirpath}` contains at least one entry that is not a file: {filepath}')

            with open(filepath, 'rb') as handle:
                try:
                    if deduplicate:
                        pseudo = pseudo_type.get_or_create(handle, filename=filename)
                    else:
                        pseudo = pseudo_type(handle, filename=filename)
                except ParsingError as exception:
                    raise ParsingError(f'failed to parse `{filepath}`: {exception}') from exception

            if pseudo.element is None:
                match = re.search(r'^([A-Za-z]{1,2})\.\w+', filename)
                if match is None:
                    raise ParsingError(
                        f'`{pseudo.__class__}` constructor did not define the element and could not parse a valid '
                        'element symbol from the filename `{filename}` either. It should have the format '
                        '`ELEMENT.EXTENSION`'
                    )
                pseudo.element = match.group(1)
            pseudos.append(pseudo)

        if not pseudos:
            raise ValueError(f'no pseudo potentials were parsed from `{dirpath}`')

        elements = set(pseudo.element for pseudo in pseudos)

        if len(pseudos) != len(elements):
            raise ValueError(f'directory `{dirpath}` contains pseudo potentials with duplicate elements')

        return pseudos

    @classmethod
    def create_from_folder(cls, dirpath, label, *, description='', pseudo_type=None, deduplicate=True):
        """Create a new ``PseudoPotentialFamily`` from the pseudo potentials contained in a directory.

        :param dirpath: absolute path to the folder containing the UPF files.
        :param label: label to give to the ``PseudoPotentialFamily``, should not already exist.
        :param description: description to give to the family.
        :param pseudo_type: subclass of ``PseudoPotentialData`` to be used for the parsed pseudos. If not specified and
            the family only defines a single supported pseudo type in ``_pseudo_types`` then that will be used otherwise
            a ``ValueError`` is raised.
        :param deduplicate: if True, will scan database for existing pseudo potentials of same type and with the same
            md5 checksum, and use that instead of the parsed one.
        :raises ValueError: if a ``PseudoPotentialFamily`` already exists with the given name.
        :raises ValueError: if ``dirpath`` is not a directory or contains anything other than files.
        :raises ValueError: if ``dirpath`` contains multiple pseudo potentials for the same element.
        :raises ValueError: if ``pseudo_type`` is explicitly specified and is not supported by this family class.
        :raises ValueError: if ``pseudo_type`` is not specified and the class supports more than one pseudo type.
        :raises ParsingError: if the constructor of the pseudo type fails for one of the files in the ``dirpath``.
        """
        type_check(description, str, allow_none=True)

        if cls.collection.count(filters={'label': label}):
            raise ValueError(f'the {cls.__name__} `{label}` already exists')

        family = cls(label=label, description=description)
        pseudos = cls.parse_pseudos_from_directory(dirpath, pseudo_type, deduplicate=deduplicate)

        # Only store the ``Group`` and the pseudo nodes now, such that we don't have to worry about the clean up in the
        # case that an exception is raised during creating them.
        family.store()
        family.add_nodes([pseudo.store() for pseudo in pseudos])

        return family

    @property
    def pseudo_type(self):
        """Return the type of the pseudopotentials that are hosted by this family.

        :return: the pseudopotential type or ``None`` if none has been set yet.
        """
        return self.base.extras.get(self._key_pseudo_type, None)

    def update_pseudo_type(self):
        """Update the pseudo type, stored as an extra, based on the current nodes in the family."""
        pseudo_types = {pseudo.__class__ for pseudo in self.pseudos.values()}

        if pseudo_types:
            assert len(pseudo_types) == 1, 'Family contains pseudopotential data nodes of various types.'
            entry_point_name = next(iter(pseudo_types)).get_entry_point_name()
        else:
            entry_point_name = None

        self.base.extras.set(self._key_pseudo_type, entry_point_name)

    def add_nodes(self, nodes):
        """Add a node or a set of nodes to the family.

        .. note: Each family instance can only contain a single pseudo potential for each element.

        :param nodes: a single or list of ``Node`` instances of type that is in ``PseudoPotentialFamily._pseudo_types``.
        :raises ModificationNotAllowed: if the family is not stored.
        :raises TypeError: if nodes are not an instance or list of instance of any of the classes listed by
            ``PseudoPotentialFamily._pseudo_types``.
        :raises ValueError: if any of the nodes are not stored or their elements already exist in this family.
        """
        if not self.is_stored:
            raise exceptions.ModificationNotAllowed('cannot add nodes to an unstored group')

        if not isinstance(nodes, (list, tuple)):
            nodes = [nodes]

        if any(not isinstance(node, self._pseudo_types) for node in nodes):
            raise TypeError(f'only nodes of types `{self._pseudo_types}` can be added: {nodes}')

        pseudos = {}

        # Check for duplicates before adding any pseudo to the internal cache
        for pseudo in nodes:
            if pseudo.element in self.elements:
                raise ValueError(f'element `{pseudo.element}` already present in this family')
            pseudos[pseudo.element] = pseudo

        self.pseudos.update(pseudos)
        self.update_pseudo_type()

        super().add_nodes(nodes)

    def remove_nodes(self, nodes):
        """Remove a pseudopotential or a set of pseudopotentials from the family.

        :param nodes: a single or list of ``PseudoPotentialData`` instances or subclasses thereof.
        """
        super().remove_nodes(nodes)

        if not isinstance(nodes, (list, tuple)):
            nodes = (nodes,)

        removed = [node.pk for node in nodes]
        self._pseudos = {pseudo.element: pseudo for pseudo in self.pseudos.values() if pseudo.pk not in removed}
        self.update_pseudo_type()

    def clear(self):
        """Remove all the pseudopotentials from this family."""
        super().clear()
        self._pseudos = None
        self.update_pseudo_type()

    @property
    def pseudos(self):
        """Return the dictionary of pseudo potentials of this family indexed on the element symbol.

        :return: dictionary of element symbol mapping pseudo potentials
        """
        if self._pseudos is None:
            self._pseudos = {pseudo.element: pseudo for pseudo in self.nodes}

        return self._pseudos

    @property
    def elements(self):
        """Return the list of elements for which this family defines a pseudo potential.

        :return: list of element symbols
        """
        return list(self.pseudos.keys())

    def get_pseudo(self, element):
        """Return the pseudo potential for the given element.

        :param element: the element for which to return the corresponding pseudo potential.
        :return: pseudo potential instance if it exists
        :raises ValueError: if the family does not contain a pseudo potential for the given element
        """
        try:
            pseudo = self.pseudos[element]
        except KeyError:
            builder = QueryBuilder()
            builder.append(self.__class__, filters={'id': self.pk}, tag='group')
            builder.append(self._pseudo_types, filters={'attributes.element': element}, with_group='group')

            try:
                pseudo = builder.one()[0]
            except exceptions.MultipleObjectsError as exception:
                raise RuntimeError(f'family `{self.label}` contains multiple pseudos for `{element}`') from exception
            except exceptions.NotExistent as exception:
                raise ValueError(
                    f'family `{self.label}` does not contain pseudo for element `{element}`'
                ) from exception
            else:
                self.pseudos[element] = pseudo

        return pseudo

    def get_pseudos(
        self,
        *,
        elements: Optional[Union[List[str], Tuple[str]]] = None,
        structure: Union[structures_classes] = None,
    ) -> Mapping[str, Union[structures_classes]]:
        """Return the mapping of kind names on pseudo potential data nodes for the given list of elements or structure.

        :param elements: list of element symbols.
        :param structure: the ``StructureData``  or ``LegacyStructureData`` node.
        :return: dictionary mapping the kind names of a structure on the corresponding pseudo potential data nodes.
        :raises ValueError: if the family does not contain a pseudo for any of the elements of the given structure.
        """
        if elements is not None and structure is not None:
            raise ValueError('cannot specify both keyword arguments `elements` and `structure`.')

        if elements is None and structure is None:
            raise ValueError('have to specify one of the keyword arguments `elements` and `structure`.')

        if elements is not None and not isinstance(elements, (list, tuple)):
            raise ValueError('elements should be a list or tuple of symbols.')

        if structure is not None and not (isinstance(structure, structures_classes)):
            raise ValueError(f'structure is of type {type(structure)} but should be of: {structures_classes}')

        if structure is not None:
            return {kind.name: self.get_pseudo(kind.symbol) for kind in structure.kinds}

        return {element: self.get_pseudo(element) for element in elements}



"""Subclass of ``PseudoPotentialFamily`` designed to represent an SSSP configuration."""
from typing import NamedTuple, Optional, Sequence

from aiida_pseudo.data.pseudo import UpfData

from ..mixins import RecommendedCutoffMixin
from .pseudo import PseudoPotentialFamily

__all__ = ('SsspConfiguration', 'SsspFamily')


class SsspConfiguration(NamedTuple):
    """Named tuple that represents an SSSP configuration."""

    version: str
    functional: str
    protocol: str

    def __str__(self):
        """Represent the configuration as a string."""
        return f'SSSP v{self.version} {self.functional} {self.protocol}'


class SsspFamily(RecommendedCutoffMixin, PseudoPotentialFamily):
    """Subclass of ``PseudoPotentialFamily`` designed to represent an SSSP configuration.

    The ``SsspFamily`` is essentially a ``PseudoPotentialFamily`` with some additional constraints. It can only be used
    to contain the pseudo potentials and corresponding metadata of an official SSSP configuration.
    """

    _pseudo_types = (UpfData,)

    label_template = 'SSSP/{version}/{functional}/{protocol}'
    filename_template = 'SSSP_{version}_{functional}_{protocol}'
    default_configuration = SsspConfiguration('1.1', 'PBE', 'efficiency')
    valid_configurations = (
        SsspConfiguration('1.0', 'PBE', 'efficiency'),
        SsspConfiguration('1.0', 'PBE', 'precision'),
        SsspConfiguration('1.1', 'PBE', 'efficiency'),
        SsspConfiguration('1.1', 'PBE', 'precision'),
        SsspConfiguration('1.1', 'PBEsol', 'efficiency'),
        SsspConfiguration('1.1', 'PBEsol', 'precision'),
        SsspConfiguration('1.2', 'PBE', 'efficiency'),
        SsspConfiguration('1.2', 'PBE', 'precision'),
        SsspConfiguration('1.2', 'PBEsol', 'efficiency'),
        SsspConfiguration('1.2', 'PBEsol', 'precision'),
        SsspConfiguration('1.3', 'PBE', 'efficiency'),
        SsspConfiguration('1.3', 'PBE', 'precision'),
        SsspConfiguration('1.3', 'PBEsol', 'efficiency'),
        SsspConfiguration('1.3', 'PBEsol', 'precision'),
    )

    @classmethod
    def get_valid_labels(cls) -> Sequence[str]:
        """Return the tuple of labels of all valid SSSP configurations."""
        return tuple(cls.format_configuration_label(configuration) for configuration in cls.valid_configurations)

    @classmethod
    def format_configuration_label(cls, configuration: SsspConfiguration) -> str:
        """Format a label for an `SsspFamily` with the required syntax.

        :param configuration: the SSSP configuration
        :return: label
        """
        return cls.label_template.format(
            version=configuration.version, functional=configuration.functional, protocol=configuration.protocol
        )

    @classmethod
    def format_configuration_filename(
        cls, configuration: SsspConfiguration, extension: str, patch_version: Optional[str] = None
    ) -> str:
        """Format the filename for a file of a particular SSSP configuration as it is available from MC Archive.

        :param configuration: the SSSP configuration.
        :param extension: the filename extension without the leading dot.
        :param patch_version: patch version of the files which overrides the ``version`` specified in the
            ``configuration``. This is necessary because we only let users specify the minor version, not install
            configurations with a specific patch version. The filename on the archive however will contain the patch
            version, so this needs to be substituted.
        :return: filename
        """
        version = configuration.version if patch_version is None else patch_version

        return (
            cls.filename_template.format(
                version=version, functional=configuration.functional, protocol=configuration.protocol
            )
            + f'.{extension}'
        )

    def __init__(self, label=None, **kwargs):
        """Construct a new instance, validating that the label matches the required format."""
        if label not in self.get_valid_labels():
            raise ValueError(f'the label `{label}` is not a valid SSSP configuration label.')

        super().__init__(label=label, **kwargs)



"""Subclass of ``PseudoPotentialFamily`` designed to represent a family with recommended cutoffs."""
from ..mixins import RecommendedCutoffMixin
from .pseudo import PseudoPotentialFamily

__all__ = ('CutoffsPseudoPotentialFamily',)


class CutoffsPseudoPotentialFamily(RecommendedCutoffMixin, PseudoPotentialFamily):
    """Subclass of ``PseudoPotentialFamily`` designed to represent a family with recommended cutoffs.

    This is mostly used for testing the functionality around the ``RecommendedCutoffMixin``.
    """



"""Command to install a pseudo potential family."""
import json
import pathlib
import sys
import tarfile
import tempfile
import typing as t

import click
import requests
import yaml
from aiida.cmdline.params import options as options_core
from aiida.cmdline.utils import decorators, echo

from .params import options, types
from .root import cmd_root

if t.TYPE_CHECKING:
    from aiida_pseudo.data.pseudo import PseudoPotentialData
    from aiida_pseudo.groups.family import PseudoDojoConfiguration, PseudoDojoFamily, SsspConfiguration


@cmd_root.group('install')
def cmd_install():
    """Install pseudo potential families."""


@cmd_install.command('family')
@click.argument('archive', type=types.PathOrUrl(exists=True, file_okay=True, path_type=pathlib.Path))
@click.argument('label', type=click.STRING)
@options_core.DESCRIPTION(help='Description for the family.')
@options.ARCHIVE_FORMAT()
@options.FAMILY_TYPE(
    type=types.PseudoPotentialFamilyTypeParam(blacklist=('pseudo.family.sssp', 'pseudo.family.pseudo_dojo'))
)
@options.PSEUDO_TYPE()
@options.TRACEBACK()
@decorators.with_dbenv()
def cmd_install_family(archive, label, description, archive_format, family_type, pseudo_type, traceback):
    """Install a standard pseudopotential family from an ARCHIVE.

    The ARCHIVE can be a (compressed) archive of a directory containing the pseudopotentials on the local file system or
    provided by an HTTP URL. Alternatively, it can be a normal directory on the local file system. The (unarchived)
    directory should only contain the pseudopotential files and they cannot be in any subdirectory. In addition,
    depending on the chosen pseudopotential type (see the option `-P/--pseudo-type`) there can be additional
    requirements on the pseudopotential file and filename format.

    If the ARCHIVE corresponds to a (compressed) archive, the command will attempt to infer the archive format from the
    filename extension of the ARCHIVE. If this fails, the archive format can be specified explicitly with the archive
    format option `-f/--archive-format`, which will also display which formats are supported. These format suffixes
    follow the naming of the `shutil.unpack_archive` standard library method.

    Once the ARCHIVE is downloaded, uncompressed and unarchived into a directory on the local file system, the command
    will create a `PseudoPotentialFamily` instance where the type of the pseudopotential data nodes that are stored
    within it is set through the `-P/--pseudo-type` option. If the default, `pseudo` (which corresponds to the data
    plugin `PseudoPotentialData`), is used, the pseudopotential files in the archive *have* to have filenames that
    strictly follow the format `ELEMENT.EXTENSION`, or the creation of the family will fail. This is because for the
    default pseudopotential type, the format of the file is unknown and the family requires the element to be known,
    which in this case can then only be parsed from the filename.

    The pseudopotential family type that is created can also be changed with the `-F/--family-type` option. Note,
    however, that not all values are accepted. For example, the `pseudo.family.sssp` and `pseudo.family.pseudo_dojo` are
    blacklisted since they have their own dedicated commands in `install sssp` and `install pseudo-dojo`, respectively.
    """
    from .utils import attempt, create_family_from_archive

    if isinstance(archive, pathlib.Path) and archive.is_dir():
        with attempt(f'creating a pseudopotential family from directory `{archive}`...', include_traceback=traceback):
            family = family_type.create_from_folder(archive, label, pseudo_type=pseudo_type)
    elif isinstance(archive, pathlib.Path) and archive.is_file():
        with attempt('unpacking archive and parsing pseudos... ', include_traceback=traceback):
            family = create_family_from_archive(
                family_type, label, archive, fmt=archive_format, pseudo_type=pseudo_type
            )
    else:
        # At this point, we can assume that it is not a valid filepath on disk, but rather a URL and the ``archive``
        # variable will contain the result objects from the ``requests`` library. The validation of the URL will already
        # have been done by the ``PathOrUrl`` parameter type and will already have retrieved the content. The content of
        # the URL must be copied to a local temporary file because `create_family_from_archive` does currently not
        # accept filelike objects, because in turn the underlying `shutil.unpack_archive` does not.
        with tempfile.NamedTemporaryFile(mode='w+b') as handle:
            handle.write(archive.content)
            handle.flush()

            with attempt('unpacking archive and parsing pseudos... ', include_traceback=traceback):
                family = create_family_from_archive(
                    family_type, label, pathlib.Path(handle.name), fmt=archive_format, pseudo_type=pseudo_type
                )

    family.description = description
    echo.echo_success(f'installed `{label}` containing {family.count()} pseudopotentials')


def download_sssp(
    configuration: 'SsspConfiguration',
    filepath_archive: pathlib.Path,
    filepath_metadata: pathlib.Path,
    traceback: bool = False,
) -> str:
    """Download the pseudopotential archive and metadata for an SSSP configuration to a path on disk.

    :param configuration: the SSSP configuration to download.
    :param filepath_archive: absolute filepath to write the pseudopotential archive to.
    :param filepath_metadata: absolute filepath to write the metadata file to.
    :param traceback: boolean, if true, print the traceback when an exception occurs.
    :return: Latest patch version of the requested minor version
    """
    from aiida_pseudo.groups.family import SsspFamily

    from .utils import attempt

    # Get the id of the latest record version from the id of the first record version
    record_id_first = 'e67tc-6b197'
    url_latest_record = f'https://archive.materialscloud.org/api/records/{record_id_first}/versions/latest'
    response = requests.get(url_latest_record)
    data = response.json()
    record_id_latest = data.get('id')

    # Url template to get a file from the latest record version
    url_template = f'https://archive.materialscloud.org/api/records/{record_id_latest}/files/{{filename}}/content'

    # Download the dictionary mapping of the minor versions to the latest corresponding patch versions. Since patch
    # releases of the SSSP only contain bug fixes, there is no reason to have the user install an outdated patch
    # version. So, the latest patch version of the minor version that is specified by the user is always installed.
    with attempt('downloading patch versions information... ', include_traceback=traceback):
        response = requests.get(url_template.format(filename='versions.yaml'), timeout=30)
        response.raise_for_status()
        # The `version_mapping` is a dictionary that maps each minor version (key) to the latest patch version (value)
        version_mapping = yaml.load(response.content, Loader=yaml.SafeLoader)
        patch_version = version_mapping[configuration.version]

    echo.echo_info(f'Latest patch version found: {patch_version}')

    archive_filename = SsspFamily.format_configuration_filename(configuration, 'tar.gz', patch_version)
    metadata_filename = SsspFamily.format_configuration_filename(configuration, 'json', patch_version)

    url_archive = url_template.format(filename=archive_filename)
    url_metadata = url_template.format(filename=metadata_filename)

    with attempt('downloading selected pseudopotentials archive... ', include_traceback=traceback):
        response = requests.get(url_archive, timeout=30)
        response.raise_for_status()
        with open(filepath_archive, 'wb') as handle:
            handle.write(response.content)
            handle.flush()

    with attempt('downloading selected pseudopotentials metadata... ', include_traceback=traceback):
        response = requests.get(url_metadata, timeout=30)
        response.raise_for_status()
        with open(filepath_metadata, 'wb') as handle:
            handle.write(response.content)
            handle.flush()

    return patch_version


def install_sssp(
    filepath_archive: pathlib.Path,
    filepath_metadata: pathlib.Path,
    label: str,
    description: str = '',
    traceback: bool = False,
) -> None:
    """Install the ``SsspFamily`` and set the recommended cutoffs.

    :param filepath_archive: path to the pseudopotential archive.
    :param filepath_metadata: path to the metadata file.
    :param label: label of the pseudopotential family group.
    :param description: description of the pseudopotential family group.
    :param traceback: boolean, if true, print the traceback when an exception occurs.
    """
    from aiida.orm import Group

    from aiida_pseudo.groups.family.sssp import SsspFamily

    from .utils import attempt, create_family_from_archive

    metadata = json.loads(filepath_metadata.read_text())

    with attempt('unpacking archive and parsing pseudos... ', include_traceback=traceback):
        family = create_family_from_archive(SsspFamily, label, filepath_archive)

    cutoffs = {}

    for element, values in metadata.items():
        if family.get_pseudo(element).md5 != values['md5']:
            Group.collection.delete(family.pk)
            msg = f"md5 of pseudo for element {element} does not match that of the metadata {values['md5']}"
            echo.echo_critical(msg)

        cutoffs[element] = {'cutoff_wfc': values['cutoff_wfc'], 'cutoff_rho': values['cutoff_rho']}

    family.description = description
    family.set_cutoffs(cutoffs, 'normal', unit='Ry')

    echo.echo_success(f'installed `{label}` containing {family.count()} pseudopotentials.')


@cmd_install.command('sssp')
@options.VERSION(type=click.Choice(['1.0', '1.1', '1.2', '1.3']), default='1.3', show_default=True)
@options.FUNCTIONAL(type=click.Choice(['PBE', 'PBEsol']), default='PBE', show_default=True)
@options.PROTOCOL(type=click.Choice(['efficiency', 'precision']), default='efficiency', show_default=True)
@options.DOWNLOAD_ONLY()
@options.FROM_DOWNLOAD()
@options.TRACEBACK()
def cmd_install_sssp(version, functional, protocol, download_only, from_download, traceback):
    """Install an SSSP configuration.

    The SSSP configuration will be automatically downloaded from the Materials Cloud Archive entry to create a new
    `SsspFamily`.
    """

    from aiida import load_profile
    from aiida.common.exceptions import ConfigurationError
    from aiida.common.files import md5_file
    from aiida.orm import QueryBuilder

    from aiida_pseudo import __version__
    from aiida_pseudo.groups.family.sssp import SsspConfiguration, SsspFamily

    if download_only and from_download is not None:
        raise click.BadParameter(
            'cannot specify both `--download-only` and `--from-download`.',
            param_hint="'--download-only' / '--from-download'",
        )

    configuration = SsspConfiguration(version, functional, protocol)
    label = SsspFamily.format_configuration_label(configuration)

    if configuration not in SsspFamily.valid_configurations:
        echo.echo_critical(f'{version} {functional} {protocol} is not a valid SSSP configuration')

    with tempfile.TemporaryDirectory() as tmppath:
        dirpath = pathlib.Path(tmppath)

        filepath_archive = dirpath / 'archive.tar.gz'
        filepath_metadata = dirpath / 'metadata.json'
        filepath_configuration = dirpath / 'configuration.json'

        if download_only:
            tarball_path = pathlib.Path.cwd() / f'{label}.aiida_pseudo'.replace('/', '_')

            if tarball_path.exists():
                echo.echo_critical(f'the file `{tarball_path}` already exists.')

            download_sssp(configuration, filepath_archive, filepath_metadata, traceback)

            with filepath_configuration.open('w') as handle:
                handle.write(json.dumps(configuration._asdict()))

            with tarfile.open(tarball_path, 'w') as tarball_file:
                for filepath in [filepath_configuration, filepath_metadata, filepath_archive]:
                    tarball_file.add(filepath, filepath.name)

            echo.echo_success(f'Pseudopotential archive written to: {tarball_path.name}')
            return

        if from_download is not None:
            tarball_path = pathlib.Path(from_download).absolute()
            with tarfile.open(tarball_path, 'r') as handle:
                handle.extractall(dirpath)

            for filepath in (filepath_archive, filepath_metadata, filepath_configuration):
                if not filepath.exists():
                    echo.echo_critical(
                        f'archive `{from_download}` did not contain required file `{filepath_archive.name}`'
                    )

            with filepath_configuration.open('r') as handle:
                configuration = SsspConfiguration(**json.load(handle))

            label = SsspFamily.format_configuration_label(configuration)

            if configuration not in SsspFamily.valid_configurations:
                echo.echo_critical(f'{version} {functional} {protocol} is not a valid SSSP configuration')

        try:
            load_profile()
        except ConfigurationError:
            echo.echo_critical(
                'Could not load a valid AiiDA profile. Check `verdi profile list` to make sure you have one defined.'
            )

        if QueryBuilder().append(SsspFamily, filters={'label': label}).first():
            echo.echo_report(f'{SsspFamily.__name__}<{label}> is already installed')
            sys.exit(1)

        if not from_download:
            download_sssp(configuration, filepath_archive, filepath_metadata, traceback)

        description = (
            f'SSSP v{configuration.version} {configuration.functional} {configuration.protocol} '
            f'installed with aiida-pseudo v{__version__}'
        )
        description += f'\nArchive pseudos md5: {md5_file(filepath_archive)}'
        description += f'\nPseudo metadata md5: {md5_file(filepath_metadata)}'

        install_sssp(filepath_archive, filepath_metadata, label, description, traceback)


def download_pseudo_dojo(
    configuration: 'PseudoDojoConfiguration',
    filepath_archive: pathlib.Path,
    filepath_metadata: pathlib.Path,
    traceback: bool = False,
) -> None:
    """Download the pseudopotential archive and metadata for a PseudoDojo configuration to a path on disk.

    :param configuration: the PseudoDojo configuration to download.
    :param filepath_archive: absolute filepath to write the pseudopotential archive to.
    :param filepath_metadata: absolute filepath to write the metadata archive to.
    :param traceback: boolean, if true, print the traceback when an exception occurs.
    """
    from ..groups.family.pseudo_dojo import PseudoDojoFamily
    from .utils import attempt

    label = PseudoDojoFamily.format_configuration_label(configuration)
    url_archive = PseudoDojoFamily.get_url_archive(label)
    url_metadata = PseudoDojoFamily.get_url_metadata(label)

    with attempt('downloading selected pseudopotentials archive... ', include_traceback=traceback):
        response = requests.get(url_archive, timeout=30, verify=False)
        response.raise_for_status()
        with open(filepath_archive, 'wb') as handle:
            handle.write(response.content)
            handle.flush()

    with attempt('downloading selected pseudopotentials metadata archive... ', include_traceback=traceback):
        response = requests.get(url_metadata, timeout=30, verify=False)
        response.raise_for_status()
        with open(filepath_metadata, 'wb') as handle:
            handle.write(response.content)
            handle.flush()


def install_pseudo_dojo(
    configuration: 'PseudoDojoConfiguration',
    filepath_archive: pathlib.Path,
    filepath_metadata: pathlib.Path,
    pseudo_type: 'PseudoPotentialData',
    label: str,
    description: str = '',
    traceback: bool = False,
) -> 'PseudoDojoFamily':
    """Install the ``PseudoDojoFamily`` and set the recommended cutoffs.

    :param filepath_archive: path to the pseudopotential archive.
    :param filepath_metadata: path to the metadata file.
    :param label: label of the pseudopotential family group.
    :pseudo_type: type of the pseudopotentials (psp8, upf, ...).
    :param description: description of the pseudopotential family group.
    :param traceback: boolean, if true, print the traceback when an exception occurs.
    """

    from aiida.orm import Group

    from aiida_pseudo.groups.family.pseudo_dojo import PseudoDojoConfiguration, PseudoDojoFamily

    from .utils import attempt, create_family_from_archive

    with attempt('unpacking archive and parsing pseudos... ', include_traceback=traceback):
        family = create_family_from_archive(PseudoDojoFamily, label, filepath_archive, pseudo_type=pseudo_type)

    with attempt('unpacking metadata archive and parsing metadata...', include_traceback=traceback):
        md5s, cutoffs = PseudoDojoFamily.parse_djrepos_from_archive(filepath_metadata, pseudo_type=pseudo_type)

    for element, md5 in md5s.items():
        if family.get_pseudo(element).md5 != md5:
            Group.collection.delete(family.pk)
            msg = f'md5 of pseudo for element {element} does not match that of the metadata {md5}'
            echo.echo_critical(msg)

    paw_configurations = (
        PseudoDojoConfiguration('1.0', 'PBE', 'SR', 'standard', 'jthxml'),
        PseudoDojoConfiguration('1.0', 'PBE', 'SR', 'stringent', 'jthxml'),
        PseudoDojoConfiguration('1.0', 'LDA', 'SR', 'standard', 'jthxml'),
        PseudoDojoConfiguration('1.0', 'LDA', 'SR', 'stringent', 'jthxml'),
    )

    # The PAW configurations have missing cutoffs for the Lanthanides, which have ben replaced with a placeholder
    # value of `-1`. We replace these with the 1.5 * the maximum cutoff from the same stringency so that these
    # potentials are still usable, but this should be taken as a _rough_ approximation.
    # We check only the `cutoff_wfc` because `cutoff_rho` is not provided by PseudoDojo and is therefore
    # locally calculated in `PseudoDojoFamily.parse_djrepos_from_archive` as `2.0 * cutoff_wfc` for PAW.
    if configuration in paw_configurations:
        adjusted_cutoffs = {}
        for stringency, str_cutoffs in cutoffs.items():
            adjusted_cutoffs[stringency] = []
            max_cutoff_wfc = max(cutoffs['cutoff_wfc'] for cutoffs in str_cutoffs.values())
            filler_cutoff_wfc = max_cutoff_wfc * 1.5
            for element, cutoff in str_cutoffs.items():
                if cutoff['cutoff_wfc'] <= 0:
                    cutoffs[stringency][element]['cutoff_wfc'] = filler_cutoff_wfc
                    cutoffs[stringency][element]['cutoff_rho'] = 2.0 * filler_cutoff_wfc
                    adjusted_cutoffs[stringency].append(element)

        for stringency, elements in adjusted_cutoffs.items():
            msg = (
                f'stringency `{stringency}` has missing recommended cutoffs for elements '
                f'{", ".join(elements)}: as a substitute, 1.5 * the maximum cutoff of the stringency '
                'was set for these elements. USE WITH CAUTION!'
            )
            echo.echo_warning(msg)

    family.description = description
    for stringency, cutoff_values in cutoffs.items():
        family.set_cutoffs(cutoff_values, stringency, unit='Eh')

    echo.echo_success(f'installed `{label}` containing {family.count()} pseudopotentials.')

    return family


@cmd_install.command('pseudo-dojo')
@options.VERSION(type=click.Choice(['0.4', '0.5', '1.0']), default='0.4', show_default=True)
@options.FUNCTIONAL(type=click.Choice(['PBE', 'PBEsol', 'LDA']), default='PBE', show_default=True)
@options.RELATIVISTIC(type=click.Choice(['SR', 'SR3plus', 'FR']), default='SR', show_default=True)
@options.PROTOCOL(type=click.Choice(['standard', 'stringent']), default='standard', show_default=True)
@options.PSEUDO_FORMAT(type=click.Choice(['psp8', 'upf', 'psml', 'jthxml']), default='psp8', show_default=True)
@options.DEFAULT_STRINGENCY(type=click.Choice(['low', 'normal', 'high']), default='normal', show_default=True)
@options.DOWNLOAD_ONLY()
@options.FROM_DOWNLOAD()
@options.TRACEBACK()
def cmd_install_pseudo_dojo(
    version,
    functional,
    relativistic,
    protocol,
    pseudo_format,
    default_stringency,
    download_only,
    from_download,
    traceback,
):
    """Install a PseudoDojo configuration.

    The PseudoDojo configuration will be automatically downloaded from pseudo-dojo.org to create a new
    `PseudoDojoFamily` subclass instance based on the specified pseudopotential format.
    """

    from aiida import load_profile
    from aiida.common.exceptions import ConfigurationError
    from aiida.common.files import md5_file
    from aiida.orm import QueryBuilder

    from aiida_pseudo import __version__
    from aiida_pseudo.data.pseudo import JthXmlData, PsmlData, Psp8Data, UpfData
    from aiida_pseudo.groups.family.pseudo_dojo import PseudoDojoConfiguration, PseudoDojoFamily

    if download_only and from_download is not None:
        raise click.BadParameter(
            'cannot specify both `--download-only` and `--from-download`.',
            param_hint="'--download-only' / '--from-download'",
        )

    pseudo_type_mapping = {
        'jthxml': JthXmlData,
        'psp8': Psp8Data,
        'psml': PsmlData,
        'upf': UpfData,
    }
    pseudo_type = pseudo_type_mapping[pseudo_format]

    configuration = PseudoDojoConfiguration(version, functional, relativistic, protocol, pseudo_format)
    label = PseudoDojoFamily.format_configuration_label(configuration)

    if configuration not in PseudoDojoFamily.valid_configurations:
        echo.echo_critical(f'{configuration} is not a valid configuration')

    with tempfile.TemporaryDirectory() as tmppath:
        dirpath = pathlib.Path(tmppath)

        filepath_archive = dirpath / 'archive.tgz'
        filepath_metadata = dirpath / 'metadata.tgz'
        filepath_configuration = dirpath / 'configuration.json'

        if download_only:
            tarball_path = pathlib.Path.cwd() / f'{label}.aiida_pseudo'.replace('/', '_')

            if tarball_path.exists():
                echo.echo_critical(f'the file `{tarball_path}` already exists.')

            download_pseudo_dojo(configuration, filepath_archive, filepath_metadata, traceback)

            with filepath_configuration.open('w') as handle:
                handle.write(json.dumps(configuration._asdict()))

            with tarfile.open(tarball_path, 'w') as tarball_file:
                for filepath in [filepath_configuration, filepath_metadata, filepath_archive]:
                    tarball_file.add(filepath, filepath.name)

            echo.echo_success(f'Pseudopotential archive written to: {tarball_path.name}')
            return

        if from_download is not None:
            tarball_path = pathlib.Path(from_download).absolute()
            with tarfile.open(tarball_path, 'r') as handle:
                handle.extractall(dirpath)

            with filepath_configuration.open('r') as handle:
                configuration = PseudoDojoConfiguration(**json.load(handle))
                pseudo_type = pseudo_type_mapping[configuration.pseudo_format]

            label = PseudoDojoFamily.format_configuration_label(configuration)

            if configuration not in PseudoDojoFamily.valid_configurations:
                echo.echo_critical(f'{configuration} is not a valid configuration')

        try:
            load_profile()
        except ConfigurationError:
            echo.echo_critical(
                'Could not load a valid AiiDA profile. Check `verdi profile list` to make sure you have one defined.'
            )

        if QueryBuilder().append(PseudoDojoFamily, filters={'label': label}).first():
            echo.echo_report(f'{PseudoDojoFamily.__name__}<{label}> is already installed')
            sys.exit(1)

        if not from_download:
            download_pseudo_dojo(configuration, filepath_archive, filepath_metadata, traceback)

        description = f'{configuration} installed with aiida-pseudo v{__version__}'
        description += f'\nArchive pseudos md5: {md5_file(filepath_archive)}'
        description += f'\nPseudo metadata md5: {md5_file(filepath_metadata)}'

        family = install_pseudo_dojo(
            configuration, filepath_archive, filepath_metadata, pseudo_type, label, description, traceback
        )
        family.set_default_stringency(default_stringency)



"""Mixin that adds support of recommended cutoffs to a ``Group`` subclass, using its extras."""
import warnings
from typing import Optional

from aiida.common.exceptions import MissingEntryPointError
from aiida.common.lang import type_check
from aiida.plugins import DataFactory

from aiida_pseudo.common.units import U

LegacyStructureData = DataFactory('core.structure')  # pylint: disable=invalid-name

try:
    StructureData = DataFactory('atomistic.structure')
except MissingEntryPointError:
    structures_classes = (LegacyStructureData,)
else:
    structures_classes = (LegacyStructureData, StructureData)

__all__ = ('RecommendedCutoffMixin',)


class RecommendedCutoffMixin:
    """Mixin that adds support of recommended cutoffs to a ``Group`` subclass, using its extras.

    This class assumes that the cutoffs apply to a plane-wave based code and as such the cutoffs pertain to the wave
    functions and the charge density. The units have to be in electronvolt.
    """

    DEFAULT_UNIT = 'eV'

    _key_cutoffs = '_cutoffs'
    _key_cutoffs_unit = '_cutoffs_unit'
    _key_default_stringency = '_default_stringency'

    @staticmethod
    def validate_cutoffs(elements: set, cutoffs: dict) -> None:
        """Validate a cutoff dictionary for a given set of elements.

        :param elements: set of elements for which to validate the cutoffs dictionary.
        :param cutoffs: dictionary with recommended cutoffs. Format: set of recommended cutoffs written as a
            dictionary that maps each element for which the family contains a pseudopotential to a dictionary that
            specifies the ``cutoff_wfc`` and ``cutoff_rho`` keys, corresponding a float value with the recommended
            cutoff to be used for the wave functions and charge density, respectively. For example:

            .. code-block::

                {
                    "Ag": {
                        "cutoff_wfc": 50.0,
                        "cutoff_rho": 200.0
                    },
                    ...
                }
        :raises ValueError: if the set of elements and those defined in the cutoffs do not match exactly, or if the
            cutoffs dictionary has an invalid format.
        """
        elements_family = set(elements)

        elements_cutoffs = set(cutoffs.keys())
        elements_diff = elements_family ^ elements_cutoffs

        if elements_family < elements_cutoffs:
            raise ValueError(f'cutoffs defined for unsupported elements: {elements_diff}')

        if elements_family > elements_cutoffs:
            raise ValueError(f'cutoffs not defined for all family elements: {elements_diff}')

        for element, values in cutoffs.items():
            if set(values.keys()) != {'cutoff_wfc', 'cutoff_rho'}:
                raise ValueError(f'invalid cutoff keys for element {element}: {values}')

            if any(not isinstance(cutoff, (int, float)) for cutoff in values.values()):
                raise ValueError(f'invalid cutoff values for element {element}: {values}')

    @staticmethod
    def validate_cutoffs_unit(unit: str) -> None:
        """Validate the cutoffs unit.

        The unit should be a name that is recognized by the ``pint`` library to be a unit of energy.

        :raises ValueError: if an invalid unit is specified.
        """
        type_check(unit, str)

        if unit not in U:
            raise ValueError(f'`{unit}` is not a valid unit.')

        if not U.Quantity(1, unit).check('[energy]'):
            raise ValueError(f'`{unit}` is not a valid energy unit.')

    def validate_stringency(self, stringency: Optional[str]) -> None:
        """Validate a cutoff stringency.

        Check if the stringency is defined for the family. If no stringency input is passed, the method checks if a
        default stringency has been set.

        :param stringency: the cutoff stringency to validate.
        :raises ValueError: if `stringency` is equal to `None` and the family defines no default stringency.
        :raises ValueError: if the family does not define cutoffs for the specified stringency.
        """
        if stringency is None:
            self.get_default_stringency()
        elif stringency not in self.get_cutoff_stringencies():
            raise ValueError(
                f'stringency `{stringency}` is not one of the available cutoff stringencies for this family: '
                f'{self.get_cutoff_stringencies()}.'
            )

    def _get_cutoffs_dict(self) -> dict:
        """Return the cutoffs dictionary that maps the stringencies to the recommended cutoffs.

        :return: the cutoffs extra or an empty dictionary if it has not yet been set.
        """
        return self.base.extras.get(self._key_cutoffs, {})

    def _get_cutoffs_unit_dict(self) -> dict:
        """Return the cutoffs units for each of the stringencies.

        :return: the cutoffs units extra or an empty dictionary if it has not yet been set.
        """
        return self.base.extras.get(self._key_cutoffs_unit, {})

    def get_default_stringency(self) -> str:
        """Return the default stringency if defined.

        :return: the default stringency.
        :raises ValueError: if default stringency has not been defined.
        """
        try:
            return self.base.extras.get(self._key_default_stringency)
        except AttributeError as exception:
            raise ValueError('no default stringency has been defined.') from exception

    def set_default_stringency(self, default_stringency: str) -> None:
        """Set the default stringency for the recommended cutoffs.

        :param default_stringency: the default stringency to be used for the recommended cutoffs.
        :raises ValueError: if the provided default stringency is not in the tuple of available cutoff stringencies for
            the pseudo family.
        :raises ValueError: if the user tries to unset the stringency by providing ``None`` as an input.
        """
        if default_stringency is None:
            raise ValueError('the default stringency cannot be unset.')

        self.validate_stringency(default_stringency)
        self.base.extras.set(self._key_default_stringency, default_stringency)

    def get_cutoff_stringencies(self) -> tuple:
        """Return a tuple of the available cutoff stringencies.

        :return: the tuple of stringencies that are defined for this family.
        """
        return tuple(self._get_cutoffs_dict().keys())

    def set_cutoffs(self, cutoffs: dict, stringency: str, unit: Optional[str] = None) -> None:
        """Set the recommended cutoffs for the pseudos in this family and a specified stringency.

        .. note: If, after the cutoffs have been set, there is only one stringency defined for the pseudo family, this
            method will automatically set this as the default. Use the ``set_default_stringency`` method to change the
            default when setting multiple stringencies.

        :param cutoffs: dictionary with recommended cutoffs. Format: set of recommended cutoffs written as a
            dictionary that maps each element for which the family contains a pseudopotential to a dictionary that
            specifies the ``cutoff_wfc`` and ``cutoff_rho`` keys, corresponding a float value with the recommended
            cutoff to be used for the wave functions and charge density, respectively. For example:

            .. code-block::

                {
                    "Ag": {
                        "cutoff_wfc": 50.0,
                        "cutoff_rho": 200.0
                    },
                    ...
                }
        :param stringency: the stringency corresponding to the provided cutoffs.
        :param unit: string definition of a unit of energy as recognized by the ``UnitRegistry`` of the ``pint`` lib.
            Defaults to electronvolt.
        :raises ValueError: if the cutoffs have an invalid format or the unit is not a valid energy unit.
        """
        unit = unit or self.DEFAULT_UNIT

        self.validate_cutoffs(set(self.elements), cutoffs)
        self.validate_cutoffs_unit(unit)

        cutoffs_dict = self._get_cutoffs_dict()
        cutoffs_dict[stringency] = cutoffs

        cutoffs_unit_dict = self._get_cutoffs_unit_dict()
        cutoffs_unit_dict[stringency] = unit

        self.base.extras.set(self._key_cutoffs, cutoffs_dict)
        self.base.extras.set(self._key_cutoffs_unit, cutoffs_unit_dict)
        if len(cutoffs_dict) == 1:
            self.set_default_stringency(stringency)

    def get_cutoffs(self, stringency: Optional[str] = None) -> dict:
        """Return a set of cutoffs for the given stringency.

        :param stringency: optional stringency for which to retrieve the cutoffs. If not specified, the default
            stringency of the family is used.
        :raises ValueError: if no stringency is specified and no default stringency is defined for the family.
        :raises ValueError: if the requested stringency is not defined for this family.
        """
        self.validate_stringency(stringency)
        stringency = stringency or self.get_default_stringency()
        return self._get_cutoffs_dict()[stringency]

    def delete_cutoffs(self, stringency: str) -> None:
        """Delete the recommended cutoffs for a specified stringency.

        .. note: If, after the cutoffs have been deleted, there is only one stringency defined for the pseudo family,
            this method will automatically set this as the default. Use the ``set_default_stringency`` method to change
            the default in case multiple stringencies are still defined.

        :param stringency: stringency for which to delete the cutoffs.
        :raises ValueError: if the requested stringency is not defined for this family.
        """
        self.validate_stringency(stringency)

        cutoffs_dict = self._get_cutoffs_dict()
        cutoffs_dict.pop(stringency)

        cutoffs_unit_dict = self._get_cutoffs_unit_dict()
        cutoffs_unit_dict.pop(stringency)

        self.base.extras.set(self._key_cutoffs, cutoffs_dict)
        self.base.extras.set(self._key_cutoffs_unit, cutoffs_unit_dict)

        warning = ''
        try:
            if stringency == self.get_default_stringency():
                self.delete_extra('_default_stringency')
                warning += f'`{stringency}` was the default stringency of this family.'
                assign_new_default = True
            else:
                assign_new_default = False
        except ValueError:
            assign_new_default = True

        if assign_new_default:
            if len(cutoffs_dict) == 0:
                warning += ' Since no other stringencies are defined for this family, no new default can be specified.'
            elif len(cutoffs_dict) == 1:
                final_stringency = next(iter(cutoffs_dict.keys()))
                self.set_default_stringency(final_stringency)
                warning += f' Setting `{final_stringency}` as the default since it is now the only defined stringency.'
            else:
                warning += (
                    f' Please set one of {self.get_cutoff_stringencies()} as the new default stringency with the '
                    '`set_default_stringency` method.'
                )

        if len(warning) > 0:
            warnings.warn(warning)

    def get_cutoffs_unit(self, stringency: Optional[str] = None) -> str:
        """Return the cutoffs unit for the specified or family default stringency.

        :param stringency: optional stringency for which to retrieve the unit. If not specified, the default stringency
            of the family is used.
        :return: the string representation of the unit of the cutoffs.
        :raises ValueError: if no stringency is specified and no default stringency is defined for the family.
        :raises ValueError: if the requested stringency is not defined for this family.
        """
        self.validate_stringency(stringency)
        stringency = stringency or self.get_default_stringency()
        return self._get_cutoffs_unit_dict()[stringency]

    def get_recommended_cutoffs(self, *, elements=None, structure=None, stringency=None, unit=None):
        """Return tuple of recommended wavefunction and density cutoffs for the given elements or ``StructureData``.

        .. note:: at least one and only one of arguments ``elements`` or ``structure`` should be passed.

        :param elements: single or tuple of elements.
        :param structure: a ``StructureData`` node.
        :param stringency: optional stringency if different from the default.
        :param unit: string definition of a unit of energy as recognized by the ``UnitRegistry`` of the ``pint`` lib.
        :return: tuple of recommended wavefunction and density cutoff.
        :raises ValueError: if the requested stringency is not defined for this family.
        :raises ValueError: if optional unit specified is invalid.
        :raises ValueError: if the family does not have a pseudo for one of the elements (of the structure).
        """
        if (elements is None and structure is None) or (elements is not None and structure is not None):
            raise ValueError('at least one and only one of `elements` or `structure` should be defined')

        type_check(elements, (tuple, str), allow_none=True)
        type_check(structure, (structures_classes), allow_none=True)

        if unit is not None:
            self.validate_cutoffs_unit(unit)

        if structure is not None:
            symbols = structure.get_symbols_set()
        elif isinstance(elements, tuple):
            symbols = elements
        else:
            symbols = (elements,)

        cutoffs_wfc = []
        cutoffs_rho = []
        cutoffs = self.get_cutoffs(stringency)

        for element in symbols:
            if element not in cutoffs:
                raise ValueError(f'family does not contain a pseudo for element `{element}`.')

            if unit is not None:
                current_unit = self.get_cutoffs_unit(stringency)
                values = {k: U.Quantity(v, current_unit).to(unit).to_tuple()[0] for k, v in cutoffs[element].items()}
            else:
                values = cutoffs[element]

            cutoffs_wfc.append(values['cutoff_wfc'])
            cutoffs_rho.append(values['cutoff_rho'])

        return (max(cutoffs_wfc), max(cutoffs_rho))



"""Base class for data types representing pseudo potentials."""
from __future__ import annotations

import io
import pathlib
import typing

from aiida import orm, plugins
from aiida.common.constants import elements
from aiida.common.exceptions import StoringNotAllowed
from aiida.common.files import md5_from_filelike
from aiida.orm.nodes.caching import NodeCaching

__all__ = ('PseudoPotentialData',)


class PseudoPotentialDataCaching(NodeCaching):
    """Class to define caching behavior of ``PseudoPotentialData`` nodes."""

    def get_objects_to_hash(self) -> dict[str, str]:
        """Return a list of objects which should be included in the node hash."""
        return {'element': self._node.element, 'md5': self._node.md5}


class PseudoPotentialData(plugins.DataFactory('core.singlefile')):
    """Base class for data types representing pseudo potentials."""

    _key_element = 'element'
    _key_md5 = 'md5'

    _CLS_NODE_CACHING = PseudoPotentialDataCaching

    @classmethod
    def get_or_create(
        cls, source: typing.Union[str, pathlib.Path, typing.BinaryIO], filename: typing.Optional[str] = None
    ):
        """Get pseudopotenial data node from database with matching md5 checksum or create a new one if not existent.

        :param source: the source pseudopotential content, either a binary stream, or a ``str`` or ``Path`` to the path
            of the file on disk, which can be relative or absolute.
        :param filename: optional explicit filename to give to the file stored in the repository.
        :return: instance of ``PseudoPotentialData``, stored if taken from database, unstored otherwise.
        :raises TypeError: if the source is not a ``str``, ``pathlib.Path`` instance or binary stream.
        :raises FileNotFoundError: if the source is a filepath but does not exist.
        """
        source = cls.prepare_source(source)

        query = orm.QueryBuilder()
        query.append(cls, subclassing=False, filters={f'attributes.{cls._key_md5}': md5_from_filelike(source)})

        pseudo = query.first(flat=True)

        if not pseudo:
            source.seek(0)
            pseudo = cls(source, filename)

        return pseudo

    @classmethod
    def get_entry_point_name(cls):
        """Return the entry point name associated with this data class.

        :return: the entry point name.
        """
        from aiida.plugins.entry_point import get_entry_point_from_class

        _, entry_point = get_entry_point_from_class(cls.__module__, cls.__name__)
        return entry_point.name

    @staticmethod
    def is_readable_byte_stream(stream) -> bool:
        """Return whether an object appears to be a readable filelike object in binary mode or stream of bytes.

        :param stream: the object to analyse.
        :returns: True if ``stream`` appears to be a readable filelike object in binary mode, False otherwise.
        """
        return isinstance(stream, io.BytesIO) or (
            hasattr(stream, 'read') and hasattr(stream, 'mode') and 'b' in stream.mode
        )

    @classmethod
    def prepare_source(cls, source: typing.Union[str, pathlib.Path, typing.BinaryIO]) -> typing.BinaryIO:
        """Validate the ``source`` representing a file on disk or a byte stream.

        .. note:: if the ``source`` is a valid file on disk, its content is read and returned as a stream of bytes.

        :raises TypeError: if the source is not a ``str``, ``pathlib.Path`` instance or binary stream.
        :raises FileNotFoundError: if the source is a filepath but does not exist.
        """
        if not isinstance(source, (str, pathlib.Path)) and not cls.is_readable_byte_stream(source):
            raise TypeError(
                f'`source` should be a `str` or `pathlib.Path` filepath on disk or a stream of bytes, got: {source}'
            )

        if isinstance(source, (str, pathlib.Path)):
            filename = pathlib.Path(source).name
            with open(source, 'rb') as handle:
                source = io.BytesIO(handle.read())
                source.name = filename

        return source

    @classmethod
    def validate_element(cls, element: str):
        """Validate the given element symbol.

        :param element: the symbol of the element following the IUPAC naming standard.
        :raises ValueError: if the element symbol is invalid.
        """
        if element not in [values['symbol'] for values in elements.values()]:
            raise ValueError(f'`{element}` is not a valid element.')

    def validate_md5(self, md5: str):
        """Validate that the md5 checksum matches that of the currently stored file.

        :param value: the md5 checksum.
        :raises ValueError: if the md5 does not match that of the currently stored file.
        """
        with self.open(mode='rb') as handle:
            md5_file = md5_from_filelike(handle)
            if md5 != md5_file:
                raise ValueError(f'md5 does not match that of stored file: {md5} != {md5_file}')

    def set_file(
        self, source: typing.Union[str, pathlib.Path, typing.BinaryIO], filename: typing.Optional[str] = None, **kwargs
    ):
        """Set the file content.

        .. note:: this method will first analyse the type of the ``source`` and if it is a filepath will convert it
            to a binary stream of the content located at that filepath, which is then passed on to the superclass. This
            needs to be done first, because it will properly set the file and filename attributes that are expected by
            other methods. Straight after the superclass call, the source seeker needs to be reset to zero if it needs
            to be read again, because the superclass most likely will have read the stream to the end. Finally it is
            important that the ``prepare_source`` is called here before the superclass invocation, because this way the
            conversion from filepath to byte stream will be performed only once. Otherwise, each subclass would perform
            the conversion over and over again.

        :param source: the source pseudopotential content, either a binary stream, or a ``str`` or ``Path`` to the path
            of the file on disk, which can be relative or absolute.
        :param filename: optional explicit filename to give to the file stored in the repository.
        :raises TypeError: if the source is not a ``str``, ``pathlib.Path`` instance or binary stream.
        :raises FileNotFoundError: if the source is a filepath but does not exist.
        """
        source = self.prepare_source(source)
        super().set_file(source, filename, **kwargs)
        source.seek(0)
        self.md5 = md5_from_filelike(source)

    def store(self, **kwargs):
        """Store the node verifying first that all required attributes are set.

        :raises :py:exc:`~aiida.common.StoringNotAllowed`: if no valid element has been defined.
        """
        try:
            self.validate_element(self.element)
        except ValueError as exception:
            raise StoringNotAllowed('no valid element has been defined.') from exception

        try:
            self.validate_md5(self.md5)
        except ValueError as exception:
            raise StoringNotAllowed(exception) from exception

        return super().store(**kwargs)

    @property
    def element(self) -> typing.Optional[int]:
        """Return the element symbol.

        :return: the symbol of the element following the IUPAC naming standard or None if not defined.
        """
        return self.base.attributes.get(self._key_element, None)

    @element.setter
    def element(self, value: str):
        """Set the element.

        :param value: the symbol of the element following the IUPAC naming standard.
        :raises ValueError: if the element symbol is invalid.
        """
        self.validate_element(value)
        self.base.attributes.set(self._key_element, value)

    @property
    def md5(self) -> typing.Optional[int]:
        """Return the md5.

        :return: the md5 of the stored file.
        """
        return self.base.attributes.get(self._key_md5, None)

    @md5.setter
    def md5(self, value: str):
        """Set the md5.

        :param value: the md5 checksum.
        :raises ValueError: if the md5 does not match that of the currently stored file.
        """
        self.validate_md5(value)
        self.base.attributes.set(self._key_md5, value)



"""Command line interface `aiida-pseudo`."""
import click
from aiida.cmdline.groups.verdi import VerdiCommandGroup

from .params import options


class CustomVerdiCommandGroup(VerdiCommandGroup):
    """Subclass of :class:`aiida.cmdline.groups.verdi.VerdiCommandGroup` for the CLI.

    This subclass overrides the verbosity option to use a custom one that removes the ``-v`` short version of the option
    since that is used by other options in this CLI and so would clash.
    """

    @staticmethod
    def add_verbosity_option(cmd):
        """Apply the ``verbosity`` option to the command, which is common to all subcommands."""
        if cmd is not None and 'verbosity' not in [param.name for param in cmd.params]:
            cmd = options.VERBOSITY()(cmd)

        return cmd


@click.group('aiida-pseudo', cls=CustomVerdiCommandGroup, context_settings={'help_option_names': ['-h', '--help']})
@options.VERBOSITY()
@options.PROFILE()
def cmd_root():
    """CLI for the ``aiida-pseudo`` plugin."""



"""Commands to inspect or modify the contents of pseudo potential families."""
import json

import click
from aiida.cmdline.params import options as options_core
from aiida.cmdline.utils import decorators, echo

from .params import arguments, options, types
from .root import cmd_root


@cmd_root.group('family')
def cmd_family():
    """Command group to inspect or modify the contents of pseudo potential families."""


@cmd_family.command('show')
@arguments.PSEUDO_POTENTIAL_FAMILY()
@options.STRINGENCY()
@options.UNIT(default=None)
@options_core.RAW()
@decorators.with_dbenv()
def cmd_family_show(family, stringency, unit, raw):
    """Show details of pseudo potential family."""
    from tabulate import tabulate

    from ..groups.mixins import RecommendedCutoffMixin

    if isinstance(family, RecommendedCutoffMixin):
        try:
            family.validate_stringency(stringency)
        except ValueError as exception:
            raise click.BadParameter(f'{exception}', param_hint="'-s' / '--stringency'")

        unit = unit or family.get_cutoffs_unit(stringency)

        headers = ['Element', 'Pseudo', 'MD5', f'Wavefunction ({unit})', f'Charge density ({unit})']
        rows = [
            [
                pseudo.element,
                pseudo.filename,
                pseudo.md5,
                *family.get_recommended_cutoffs(elements=pseudo.element, stringency=stringency, unit=unit),
            ]
            for pseudo in family.nodes
        ]
    else:
        headers = ['Element', 'Pseudo', 'MD5']
        rows = [[pseudo.element, pseudo.filename, pseudo.md5] for pseudo in family.nodes]

    if raw:
        echo.echo(tabulate(sorted(rows), tablefmt='plain', floatfmt='.1f'))
    else:
        echo.echo(tabulate(sorted(rows), headers=headers, floatfmt='.1f'))


@cmd_family.group('cutoffs')
def cmd_family_cutoffs():
    """Command group to inspect or modify the recommended cutoffs of pseudo potential families."""


@cmd_family_cutoffs.command('set')
@arguments.PSEUDO_POTENTIAL_FAMILY(
    type=types.PseudoPotentialFamilyParam(blacklist=('pseudo.family.sssp', 'pseudo.family.pseudo_dojo'))
)
@click.argument('cutoffs', type=click.File(mode='rb'))
@options.STRINGENCY(required=True)
@options.UNIT()
@decorators.with_dbenv()
def cmd_family_cutoffs_set(family, cutoffs, stringency, unit):
    """Set the recommended cutoffs for a pseudo potential family and a specified stringency.

    The cutoffs should be provided as a JSON file through the argument `CUTOFFS` which should have the structure:

    \b
        {
            "Ag": {
                "cutoff_wfc": 50.0,
                "cutoff_rho": 200.0
            },
            ...
        }

    where the cutoffs are expected to be in electronvolt by default.
    """
    from ..groups.mixins import RecommendedCutoffMixin

    if not isinstance(family, RecommendedCutoffMixin):
        raise click.BadParameter(f'family `{family}` does not support recommended cutoffs to be set.')

    try:
        data = json.load(cutoffs)
    except ValueError as exception:
        raise click.BadParameter(f'`{cutoffs.name}` contains invalid JSON: {exception}', param_hint='CUTOFFS')

    cutoffs_dict = {}
    for element, values in data.items():
        try:
            cutoffs_dict[element] = {'cutoff_wfc': values['cutoff_wfc'], 'cutoff_rho': values['cutoff_rho']}
        except KeyError as exception:
            raise click.BadParameter(
                f'`{cutoffs.name}` is missing cutoffs for element `{element}`: {exception}', param_hint='CUTOFFS'
            ) from exception

    try:
        family.set_cutoffs(cutoffs_dict, stringency, unit=unit)
    except ValueError as exception:
        raise click.BadParameter(f'`{cutoffs.name}` contains invalid cutoffs: {exception}', param_hint='CUTOFFS')

    echo.echo_success(f'set cutoffs for `{family}` with the stringency `{stringency}`.')



"""Reusable options for CLI commands."""
import functools
import shutil
from pathlib import Path

import click
from aiida.cmdline.params import options as core_options
from aiida.cmdline.params import types as core_types

from .types import PseudoPotentialFamilyTypeParam, PseudoPotentialTypeParam, UnitParamType

__all__ = (
    'PROFILE',
    'VERBOSITY',
    'VERSION',
    'FUNCTIONAL',
    'RELATIVISTIC',
    'PROTOCOL',
    'PSEUDO_FORMAT',
    'STRINGENCY',
    'DEFAULT_STRINGENCY',
    'TRACEBACK',
    'FAMILY_TYPE',
    'ARCHIVE_FORMAT',
    'UNIT',
    'DOWNLOAD_ONLY',
)

PROFILE = functools.partial(
    core_options.PROFILE, type=core_types.ProfileParamType(load_profile=True), expose_value=False
)

# Clone the ``VERBOSITY`` option from ``aiida-core`` so the ``-v`` short flag can be removed, since that overlaps with
# the flag of the ``VERSION`` option of this CLI.
VERBOSITY = core_options.VERBOSITY.clone()
VERBOSITY.args = ('--verbosity',)

VERSION = core_options.OverridableOption(
    '-v', '--version', type=click.STRING, required=False, help='Select the version of the installed configuration.'
)

FUNCTIONAL = core_options.OverridableOption(
    '-x',
    '--functional',
    type=click.STRING,
    required=False,
    help='Select the functional of the installed configuration.',
)

RELATIVISTIC = core_options.OverridableOption(
    '-r',
    '--relativistic',
    type=click.STRING,
    required=False,
    help='Select the type of relativistic effects included in the installed configuration.',
)

PROTOCOL = core_options.OverridableOption(
    '-p', '--protocol', type=click.STRING, required=False, help='Select the protocol of the installed configuration.'
)

PSEUDO_FORMAT = core_options.OverridableOption(
    '-f',
    '--pseudo-format',
    type=click.STRING,
    required=True,
    help='Select the pseudopotential file format of the installed configuration.',
)

STRINGENCY = core_options.OverridableOption(
    '-s', '--stringency', type=click.STRING, required=False, help='Stringency level for the recommended cutoffs.'
)

DEFAULT_STRINGENCY = core_options.OverridableOption(
    '-s',
    '--default-stringency',
    type=click.STRING,
    required=False,
    help='Select the default stringency level for the installed configuration.',
)

TRACEBACK = core_options.OverridableOption(
    '-t', '--traceback', is_flag=True, help='Include the stacktrace if an exception is encountered.'
)

FAMILY_TYPE = core_options.OverridableOption(
    '-F',
    '--family-type',
    type=PseudoPotentialFamilyTypeParam(),
    default='pseudo.family',
    show_default=True,
    help='Choose the type of pseudo potential family to create.',
)

PSEUDO_TYPE = core_options.OverridableOption(
    '-P',
    '--pseudo-type',
    type=PseudoPotentialTypeParam(),
    default='pseudo',
    show_default=True,
    help=(
        'Select the pseudopotential type to be used for the family. Should be the entry point name of a '
        'subclass of `PseudoPotentialData`.'
    ),
)

ARCHIVE_FORMAT = core_options.OverridableOption(
    '-f', '--archive-format', type=click.Choice([fmt[0] for fmt in shutil.get_archive_formats()])
)

UNIT = core_options.OverridableOption(
    '-u',
    '--unit',
    type=UnitParamType(quantity='energy'),
    required=False,
    default='eV',
    show_default=True,
    help='Specify the energy unit of the cutoffs. Must be recognized by the ``UnitRegistry`` of the ``pint`` library.',
)

DOWNLOAD_ONLY = core_options.OverridableOption(
    '--download-only',
    is_flag=True,
    help=(
        'Only download the pseudopotential files to the current working directory, without installing the '
        'pseudopotential family.'
    ),
)

FROM_DOWNLOAD = core_options.OverridableOption(
    '--from-download',
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    required=False,
    help='Install the pseudpotential family from archive and metadata downloaded with the `--download-only` option.',
)



"""Module for the command line interface."""
from .family import cmd_family  # noqa: F401
from .install import cmd_install, cmd_install_family, cmd_install_pseudo_dojo, cmd_install_sssp  # noqa: F401
from .list import cmd_list  # noqa: F401
from .root import cmd_root  # noqa: F401



"""Module with data plugins to represent pseudo potentials."""
from .jthxml import *
from .pseudo import *
from .psf import *
from .psml import *
from .psp8 import *
from .upf import *
from .vps import *

__all__ = jthxml.__all__ + pseudo.__all__ + psf.__all__ + psml.__all__ + psp8.__all__ + upf.__all__ + vps.__all__



"""Commands to list instances of `PseudoPotentialFamily`."""
import click
from aiida.cmdline.params import options as options_core
from aiida.cmdline.utils import decorators, echo

from .params import options
from .root import cmd_root

PROJECTIONS_VALID = ('pk', 'uuid', 'type_string', 'label', 'description', 'count')
PROJECTIONS_DEFAULT = ('label', 'type_string', 'count')


def get_families_builder():
    """Return a query builder that will query for instances of `PseudoPotentialFamily` or its subclasses.

    :return: `QueryBuilder` instance
    """
    from aiida.orm import QueryBuilder

    from aiida_pseudo.groups.family import PseudoPotentialFamily

    builder = QueryBuilder().append(PseudoPotentialFamily)

    return builder


@cmd_root.command('list')
@options_core.PROJECT(type=click.Choice(PROJECTIONS_VALID), default=PROJECTIONS_DEFAULT)
@options_core.RAW()
@options.FAMILY_TYPE(default=None, help='Filter for families of with this type string.')
@decorators.with_dbenv()
def cmd_list(project, raw, family_type):
    """List installed pseudo potential families."""
    from tabulate import tabulate

    mapping_project = {
        'count': lambda family: family.count(),
    }

    if get_families_builder().count() == 0:
        echo.echo_report('no pseudo potential families have been installed yet: use `aiida-pseudo install`.')
        return

    rows = []

    for (group,) in get_families_builder().iterall():
        if family_type and family_type.entry_point != group.type_string:
            continue

        row = []

        for projection in project:
            try:
                projected = mapping_project[projection](group)
            except KeyError:
                projected = getattr(group, projection)
            row.append(projected)

        rows.append(row)

    if not rows:
        echo.echo_report('no pseudo potential families found that match the filtering criteria.')
        return

    if raw:
        echo.echo(tabulate(rows, disable_numparse=True, tablefmt='plain'))
    else:
        headers = [projection.replace('_', ' ').capitalize() for projection in project]
        echo.echo(tabulate(rows, headers=headers, disable_numparse=True))



"""Module for data plugin to represent a pseudo potential in VPS format."""
import pathlib
import re
import typing

from aiida.common.constants import elements

from .pseudo import PseudoPotentialData

__all__ = ('VpsData',)

PATTERN_FLOAT = r'[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?'
REGEX_ATOMIC_NUMBER = re.compile(r"""\s*AtomSpecies\s*(?P<atomic_number>[\d]{1,3})\s*""", re.I)
REGEX_Z_VALENCE = re.compile(r"""\s*valence\.electron\s*(?P<z_valence>""" + PATTERN_FLOAT + r""")\s*""", re.I)
REGEX_XC_TYPE = re.compile(r"""\s*xc\.type\s*(?P<xc_type>[A-Z]{3})\s*""", re.I)

VALID_XC_TYPES = ('LDA', 'LSDA-CA', 'LSDA-PW', 'GGA-PBE', 'EXX-TEST')


def parse_element(content: str):
    """Parse the content of the VPS file to determine the element.

    :param content: the decoded content of the file.
    :return: the symbol of the element following the IUPAC naming standard.
    """
    match = REGEX_ATOMIC_NUMBER.search(content)

    if match:
        atomic_number = match.group('atomic_number')

        try:
            atomic_number = int(atomic_number)
        except ValueError as exception:
            raise ValueError(
                f'parsed value for the atomic number `{atomic_number}` is not a valid number.'
            ) from exception

        try:
            element = elements[atomic_number]['symbol']
        except KeyError as exception:
            raise ValueError(
                f'parsed value for the atomic number `{atomic_number}` is not in aiida.common.constants.elements.'
            ) from exception

        return element

    raise ValueError(f'could not parse the element from the VPS content: {content}')


def parse_z_valence(content: str):
    """Parse the content of the VPS file to determine the Z valence.

    :param content: the decoded content of the file.
    :return: the number of valence electrons for which the pseudopotential was generated.
    """
    match = REGEX_Z_VALENCE.search(content)

    if match:
        z_valence = match.group('z_valence')

        try:
            z_valence = float(z_valence)
        except ValueError as exception:
            raise ValueError(f'parsed value for the Z valence `{z_valence}` is not a valid number.') from exception

        if int(z_valence) != z_valence:
            raise ValueError(f'parsed value for the Z valence `{z_valence}` is not an integer')

        return int(z_valence)

    raise ValueError(f'could not parse the Z valence from the VPS content: {content}')


def parse_xc_type(content: str):
    """Parse the content of the VPS file to determine the exchange-correlation functional type.

    :param content: the decoded content of the file.
    :return: the OpenMX-compatible name of the exchange-correlation type.
    """
    match = REGEX_XC_TYPE.search(content)

    if match:
        xc_type = match.group('xc_type')

        if xc_type == 'GGA':
            xc_type = 'GGA-PBE'

        if xc_type not in VALID_XC_TYPES:
            raise ValueError(
                f'parsed value for the exchange-correlation type `{xc_type}` is not a valid OpenMX XcType string.'
            )

        return xc_type

    raise ValueError(f'could not parse the exchange-correlation type from the VPS content: {content}')


class VpsData(PseudoPotentialData):
    """Data plugin to represent a pseudo potential in VPS format."""

    _key_z_valence = 'z_valence'
    _key_xc_type = 'xc_type'

    def set_file(
        self, source: typing.Union[str, pathlib.Path, typing.BinaryIO], filename: typing.Optional[str] = None, **kwargs
    ):
        """Set the file content and parse other optional attributes from the content.

        .. note:: this method will first analyse the type of the ``source`` and if it is a filepath will convert it
            to a binary stream of the content located at that filepath, which is then passed on to the superclass. This
            needs to be done first, because it will properly set the file and filename attributes that are expected by
            other methods. Straight after the superclass call, the source seeker needs to be reset to zero if it needs
            to be read again, because the superclass most likely will have read the stream to the end. Finally it is
            important that the ``prepare_source`` is called here before the superclass invocation, because this way the
            conversion from filepath to byte stream will be performed only once. Otherwise, each subclass would perform
            the conversion over and over again.

        :param source: the source pseudopotential content, either a binary stream, or a ``str`` or ``Path`` to the path
            of the file on disk, which can be relative or absolute.
        :param filename: optional explicit filename to give to the file stored in the repository.
        :raises TypeError: if the source is not a ``str``, ``pathlib.Path`` instance or binary stream.
        :raises FileNotFoundError: if the source is a filepath but does not exist.
        :raises ValueError: if the element symbol is invalid.
        """
        source = self.prepare_source(source)
        super().set_file(source, filename, **kwargs)
        source.seek(0)
        content = source.read().decode('utf-8')
        self.element = parse_element(content)
        self.z_valence = parse_z_valence(content)
        self.xc_type = parse_xc_type(content)

    @property
    def z_valence(self) -> typing.Optional[int]:
        """Return the Z valence.

        :return: the Z valence.
        """
        return self.base.attributes.get(self._key_z_valence, None)

    @z_valence.setter
    def z_valence(self, value: int):
        """Set the Z valence.

        :param value: the Z valence.
        :raises ValueError: if the value is not a positive integer
        """
        if not isinstance(value, int) or value < 0:
            raise ValueError(f'`{value}` is not a positive integer.')

        self.base.attributes.set(self._key_z_valence, value)

    @property
    def xc_type(self) -> typing.Optional[int]:
        """Return the exchange-correlation type.

        :return: the exchange-correlation type.
        """
        return self.base.attributes.get(self._key_xc_type, None)

    @xc_type.setter
    def xc_type(self, value: str):
        """Set the exchange-correlation type.

        :param value: the exchange-correlation type.
        :raises ValueError: if the value is not a valid OpenMX XcType
        """
        if not isinstance(value, str) or value not in VALID_XC_TYPES:
            raise ValueError(f'`{value}` is not a valid OpenMX XcType string.')

        self.base.attributes.set(self._key_xc_type, value)



"""Reusable arguments for CLI commands."""
from aiida.cmdline.params.arguments.overridable import OverridableArgument

from .types import PseudoPotentialFamilyParam

__all__ = ('PSEUDO_POTENTIAL_FAMILY',)

PSEUDO_POTENTIAL_FAMILY = OverridableArgument(
    'family', type=PseudoPotentialFamilyParam(sub_classes=('aiida.groups:pseudo.family',))
)
